"""Shaft Element module.

This module defines the ShaftElement classes which will be used to represent the rotor
shaft. There're 2 options, an element with 8 or 12 degrees of freedom.
"""

import inspect
import os
from pathlib import Path

import numpy as np
import toml
from plotly import graph_objects as go

from ross.element import Element
from ross.materials import Material, steel
from ross.units import Q_, check_units
from ross.utils import read_table_file

__all__ = ["ShaftElement", "ShaftElement6DoF"]


class ShaftElement(Element):
    r"""A shaft element.

    This class will create a shaft element that may take into
    account shear, rotary inertia an gyroscopic effects. The object can be
    cylindrical or conical and the formulation is based on :cite:`genta1988conical`
    The matrices will be defined considering the following local
    coordinate vector:

    .. math::

        [x_0, y_0, \alpha_0, \beta_0, x_1, y_1, \alpha_1, \beta_1]^T
    Where :math:`\alpha_0` and :math:`\alpha_1` are the bending on the yz plane
    and :math:`\beta_0` and :math:`\beta_1` are the bending on the xz plane.

    Parameters
    ----------
    L : float, pint.Quantity
        Element length (m).
    idl : float, pint.Quantity
        Inner diameter of the element at the left position (m).
    odl : float, pint.Quantity
        Outer diameter of the element at the left position (m).
    idr : float, pint.Quantity, optional
        Inner diameter of the element at the right position (m).
        Default is equal to idl value (cylindrical element).
    odr : float, pint.Quantity, optional
        Outer diameter of the element at the right position (m).
        Default is equal to odl value (cylindrical element).
    material : ross.Material
        Shaft material.
    n : int, optional
        Element number (coincident with it's first node).
        If not given, it will be set when the rotor is assembled
        according to the element's position in the list supplied to
        the rotor constructor.
    axial_force : float, optional
        Axial force (N).
    torque : float, optional
        Torque (N*m).
    shear_effects : bool, optional
        Determine if shear effects are taken into account.
        Default is True.
    rotary_inertia : bool, optional
        Determine if rotary_inertia effects are taken into account.
        Default is True.
    gyroscopic : bool, optional
        Determine if gyroscopic effects are taken into account.
        Default is True.
    shear_method_calc : str, optional
        Determines which shear calculation method the user will adopt
        Default is 'cowper'
    alpha : float, optional
        Mass proportional damping factor.
        Default is zero.
    beta : float, optional
        Stiffness proportional damping factor.
        Default is zero.
    tag : str, optional
        Element tag.
        Default is None.

    Returns
    -------
    shaft_element : ross.ShaftElement
        A shaft_element object.

    Attributes
    ----------
    Poisson : float
        Poisson coefficient for the element.
    A : float
        Element section area at half length (m**2).
    A_l : float
        Element section area at left end (m**2).
    A_r : float
        Element section area at right end (m**2).
    beam_cg : float
        Element center of gravity local position (m).
    axial_cg_pos : float
        Element center of gravity global position (m).
        This should be used only after the rotor is built.
        Default is None.
    Ie : float
        Ie is the second moment of area of the cross section about
        the neutral plane (m**4).
    phi : float
        Constant that is used according to :cite:`friswell2010dynamics` to
        consider rotary inertia and shear effects. If these are not considered
        :math:`\phi=0`.
    kappa : float
        Shear coefficient for the element.

    References
    ----------
    .. bibliography::
        :filter: docname in docnames

    Examples
    --------
    >>> from ross.materials import steel
    >>> # Euler-Bernoulli conical element
    >>> Euler_Bernoulli_Element = ShaftElement(
    ...                         material=steel, L=0.5, idl=0.05, odl=0.1,
    ...                         idr=0.05, odr=0.15,
    ...                         rotary_inertia=False,
    ...                         shear_effects=False)
    >>> Euler_Bernoulli_Element.phi
    0
    >>> # Timoshenko cylindrical element. In this case idr and odr are omitted.
    >>> Timoshenko_Element = ShaftElement(
    ...                         material=steel, L=0.5, idl=0.05, odl=0.1,
    ...                         rotary_inertia=True,
    ...                         shear_effects=True)
    >>> Timoshenko_Element.phi
    0.1571268472906404
    """

    @check_units
    def __init__(
        self,
        L,
        idl,
        odl,
        idr=None,
        odr=None,
        material=None,
        n=None,
        axial_force=0,
        torque=0,
        shear_effects=True,
        rotary_inertia=True,
        gyroscopic=True,
        shear_method_calc="cowper",
        tag=None,
        alpha=0,
        beta=0,
    ):
        if idr is None:
            idr = idl
        if odr is None:
            odr = odl

        if material is None:
            raise AttributeError("Material is not defined.")

        if type(material) is str:
            self.material = Material.load_material(material)
        else:
            self.material = material

        self.shear_effects = shear_effects
        self.rotary_inertia = rotary_inertia
        self.gyroscopic = gyroscopic
        self.axial_force = axial_force
        self.torque = torque
        self._n = n
        self.n_l = n
        self.n_r = None
        if n is not None:
            self.n_r = n + 1

        self.tag = tag
        self.shear_method_calc = shear_method_calc

        self.L = float(L)
        self.o_d = (float(odl) + float(odr)) / 2
        self.i_d = (float(idl) + float(idr)) / 2
        self.idl = float(idl)
        self.odl = float(odl)
        self.idr = float(idr)
        self.odr = float(odr)
        self.color = self.material.color

        self.alpha = float(alpha)
        self.beta = float(beta)

        # A_l = cross section area from the left side of the element
        # A_r = cross section area from the right side of the element
        A_l = np.pi * (odl**2 - idl**2) / 4
        A_r = np.pi * (odr**2 - idr**2) / 4
        self.A_l = A_l
        self.A_r = A_r

        # Second moment of area of the cross section from the left side
        # of the element
        Ie_l = np.pi * (odl**4 - idl**4) / 64
        Ie_r = np.pi * (odr**4 - idr**4) / 64
        self.Ie_l = Ie_l
        self.Ie_r = Ie_r

        outer = self.odl**2 + self.odl * self.odr + self.odr**2
        inner = self.idl**2 + self.idl * self.idr + self.idr**2
        self.volume = np.pi * (self.L / 12) * (outer - inner)
        self.m = self.material.rho * self.volume

        roj = odl / 2
        rij = idl / 2
        rok = odr / 2
        rik = idr / 2

        # geometrical coefficients
        delta_ro = rok - roj
        delta_ri = rik - rij
        a1 = 2 * np.pi * (roj * delta_ro - rij * delta_ri) / A_l
        a2 = np.pi * (roj**3 * delta_ro - rij**3 * delta_ri) / Ie_l
        b1 = np.pi * (delta_ro**2 - delta_ri**2) / A_l
        b2 = 3 * np.pi * (roj**2 * delta_ro**2 - rij**2 * delta_ri**2) / (2 * Ie_l)
        gama = np.pi * (roj * delta_ro**3 - rij * delta_ri**3) / Ie_l
        delta = np.pi * (delta_ro**4 - delta_ri**4) / (4 * Ie_l)

        self.a1 = a1
        self.a2 = a2
        self.b1 = b1
        self.b2 = b2
        self.gama = gama
        self.delta = delta

        # the area is calculated from the cross section located in the middle
        # of the element
        self.A = A_l * (1 + a1 * 0.5 + b1 * 0.5**2)

        # Ie is the second moment of area of the cross section - located in
        # the middle of the element - about the neutral plane
        Ie = Ie_l * (1 + a2 * 0.5 + b2 * 0.5**2 + gama * 0.5**3 + delta * 0.5**4)
        self.Ie = Ie

        phi = 0

        # geometric center
        c1 = roj**2 + 2 * roj * rok + 3 * rok**2 - rij**2 - 2 * rij * rik - 3 * rik**2
        c2 = (roj**2 + roj * rok + rok**2) - (rij**2 + rij * rik + rik**2)
        self.beam_cg = L * c1 / (4 * c2)
        self.axial_cg_pos = None

        # Slenderness ratio of beam elements (G*A*L**2) / (E*I)
        sld = (self.material.G_s * self.A * self.L**2) / (self.material.E * Ie)
        self.slenderness_ratio = sld

        # Moment of inertia
        # fmt: off
        self.Im = (
            (np.pi * L * (self.m / self.volume) / 10) *
            ((roj ** 4 + roj ** 3 * rok + roj ** 2 * rok ** 2 + roj * rok ** 3 + rok ** 4) -
             (rij ** 4 + rij ** 3 * rik + rij ** 2 * rik ** 2 + rij * rik ** 3 + rik ** 4))
        )
        # fmt: on

        # picking a method to calculate the shear coefficient
        # List of avaible methods:
        # hutchinson - kappa as per Hutchinson (2001)
        # cowper - kappa as per Cowper (1996)
        if shear_effects:
            r = ((idl + idr) / 2) / ((odl + odr) / 2)
            r2 = r * r
            r12 = (1 + r2) ** 2
            if shear_method_calc == "hutchinson":
                # Shear coefficient (phi)
                # kappa as per Hutchinson (2001)
                # fmt: off
                kappa = 6 * r12 * ((1 + self.material.Poisson) /
                        ((r12 * (7 + 12 * self.material.Poisson + 4 * self.material.Poisson ** 2) +
                        4 * r2 * (5 + 6 * self.material.Poisson + 2 * self.material.Poisson ** 2))))
                # fmt: on
            elif shear_method_calc == "cowper":
                # kappa as per Cowper (1996)
                # fmt: off
                kappa = 6 * r12 * (
                    (1 + self.material.Poisson)
                    / (r12 * (7 + 6 * self.material.Poisson) + r2 * (20 + 12 * self.material.Poisson))
                )
                # fmt: on
            else:
                raise Warning(
                    "This method of calculating shear coefficients is not implemented. See guide for futher informations."
                )

            # fmt: off
            phi = 12 * self.material.E * self.Ie / (self.material.G_s * kappa * self.A * L ** 2)
            # fmt: on
            self.kappa = kappa

        self.phi = phi
        self.dof_global_index = None

    def __eq__(self, other):
        """Equality method for comparasions.

        Parameters
        ----------
        other : obj
            parameter for comparasion

        Returns
        -------
        bool
            True if the comparison is true; False otherwise.

        Example
        -------
        >>> from ross.materials import steel
        >>> shaft1 = ShaftElement(
        ...        L=0.25, idl=0, idr=0, odl=0.05, odr=0.08,
        ...        material=steel, rotary_inertia=True, shear_effects=True
        ... )
        >>> shaft2 = ShaftElement(
        ...        L=0.25, idl=0, idr=0, odl=0.05, odr=0.08,
        ...        material=steel, rotary_inertia=True, shear_effects=True
        ... )
        >>> shaft1 == shaft2
        True
        """
        if self.__dict__ == other.__dict__:
            return True
        else:
            return False

    def __repr__(self):
        """Return a string representation of a shaft element.

        Returns
        -------
        A string representation of a shaft element object.

        Examples
        --------
        >>> from ross.materials import steel
        >>> shaft1 = ShaftElement(
        ...        L=0.25, idl=0, idr=0, odl=0.05, odr=0.08,
        ...        material=steel, rotary_inertia=True, shear_effects=True
        ... )
        >>> shaft1 # doctest: +ELLIPSIS
        ShaftElement(L=0.25, idl=0.0...
        """
        return (
            f"{self.__class__.__name__}"
            f"(L={self.L:{0}.{5}}, idl={self.idl:{0}.{5}}, "
            f"idr={self.idr:{0}.{5}}, odl={self.odl:{0}.{5}},  "
            f"odr={self.odr:{0}.{5}}, material={self.material.name!r}, "
            f"n={self.n})"
        )

    def __str__(self):
        """Convert object into string.

        Returns
        -------
        The object's parameters translated to strings

        Example
        -------
        >>> print(ShaftElement(L=0.25, idl=0, odl=0.05, odr=0.08, material=steel))
        Element Number:             None
        Element Lenght   (m):       0.25
        Left Int. Diam.  (m):        0.0
        Left Out. Diam.  (m):       0.05
        Right Int. Diam. (m):        0.0
        Right Out. Diam. (m):       0.08
        -----------------------------------
        Steel
        -----------------------------------
        Density         (kg/m**3): 7810.0
        Young`s modulus (N/m**2):  2.11e+11
        Shear modulus   (N/m**2):  8.12e+10
        Poisson coefficient     :  0.29926108
        """
        return (
            f"Element Number:             {self.n}"
            f"\nElement Lenght   (m): {self.L:{10}.{5}}"
            f"\nLeft Int. Diam.  (m): {self.idl:{10}.{5}}"
            f"\nLeft Out. Diam.  (m): {self.odl:{10}.{5}}"
            f"\nRight Int. Diam. (m): {self.idr:{10}.{5}}"
            f"\nRight Out. Diam. (m): {self.odr:{10}.{5}}"
            f'\n{35*"-"}'
            f"\n{self.material}"
        )

    def __hash__(self):
        return hash(self.tag)

    def save(self, file):
        signature = inspect.signature(self.__init__)
        args_list = list(signature.parameters)
        args = {arg: getattr(self, arg) for arg in args_list}

        # add material characteristics so that the shaft element can be reconstructed
        # even if the material is not in the available_materials file.
        args["material"] = {
            "name": self.material.name,
            "rho": self.material.rho,
            "E": self.material.E,
            "G_s": self.material.G_s,
            "color": self.material.color,
        }

        try:
            data = toml.load(file)
        except FileNotFoundError:
            data = {}

        data[f"{self.__class__.__name__}_{self.tag}"] = args
        with open(file, "w") as f:
            toml.dump(data, f)

    @classmethod
    def read_toml_data(cls, data):
        data["material"] = Material(**data["material"])
        return cls(**data)

    @property
    def n(self):
        """Set the element number as property.

        Returns
        -------
        n : int
            Element number
        """
        return self._n

    @n.setter
    def n(self, value):
        """Set a new value for the element number.

        Parameters
        ----------
        value : int
            element number

        Example
        -------
        >>> from ross.materials import steel
        >>> shaft1 = ShaftElement(
        ...        L=0.25, idl=0, idr=0, odl=0.05, odr=0.08,
        ...        material=steel, rotary_inertia=True, shear_effects=True
        ... )
        >>> shaft1.n = 0
        >>> shaft1 # doctest: +ELLIPSIS
        ShaftElement(L=0.25, idl=0.0...
        """
        self._n = value
        self.n_l = value
        if value is not None:
            self.n_r = value + 1

    def dof_mapping(self):
        """Degrees of freedom mapping.

        Returns a dictionary with a mapping between degree of freedom and its index.

        Returns
        -------
        dof_mapping : dict
            A dictionary containing the degrees of freedom and their indexes.

        Examples
        --------
        The numbering of the degrees of freedom for each node.

        Being the following their ordering for a node:

        x_0 - horizontal translation
        y_0 - vertical translation
        z_0 - axial translation
        alpha_0 - rotation around horizontal
        beta_0  - rotation around vertical
        theta_0 - torsion around axial

        >>> sh = ShaftElement(L=0.5, idl=0.05, odl=0.1, material=steel,
        ...                   rotary_inertia=True, shear_effects=True)
        >>> sh.dof_mapping()["x_0"]
        0
        """
        return dict(
            x_0=0, y_0=1, alpha_0=2, beta_0=3, x_1=4, y_1=5, alpha_1=6, beta_1=7
        )

    def M(self):
        """Mass matrix for an instance of a shaft element.

        Returns
        -------
        M : np.ndarray
            Mass matrix for the shaft element.

        Examples
        --------
        >>> Timoshenko_Element = ShaftElement(
        ...                         L=0.5, idl=0.05, idr=0.05, odl=0.1,
        ...                         odr=0.15, material=steel,
        ...                         rotary_inertia=True,
        ...                         shear_effects=True)
        >>> Timoshenko_Element.M()[:4, :4]
        array([[11.36986417,  0.        ,  0.        ,  0.86197637],
               [ 0.        , 11.36986417, -0.86197637,  0.        ],
               [ 0.        , -0.86197637,  0.08667495,  0.        ],
               [ 0.86197637,  0.        ,  0.        ,  0.08667495]])
        """
        phi = self.phi
        L = self.L
        a1 = self.a1
        a2 = self.a2
        b1 = self.b1
        b2 = self.b2
        delta = self.delta
        gama = self.gama
        Ie_l = self.Ie_l
        A_l = self.A_l

        m1 = (
            (468 + 882 * phi + 420 * phi**2)
            + a1 * (108 + 210 * phi + 105 * phi**2)
            + b1 * (38 + 78 * phi + 42 * phi**2)
        )
        m2 = (
            (66 + 115.5 * phi + 52.5 * phi**2)
            + a1 * (21 + 40.5 * phi + 21 * phi**2)
            + b1 * (8.5 + 18 * phi + 10.5 * phi**2)
        )
        m3 = (
            (162 + 378 * phi + 210 * phi**2)
            + a1 * (81 + 189 * phi + 105 * phi**2)
            + b1 * (46 + 111 * phi + 63 * phi**2)
        )
        m4 = (
            (39 + 94.5 * phi + 52.5 * phi**2)
            + a1 * (18 + 40.5 * phi + 21 * phi**2)
            + b1 * (9.5 + 21 * phi + 10.5 * phi**2)
        )
        m5 = (
            (12 + 21 * phi + 10.5 * phi**2)
            + a1 * (4.5 + 9 * phi + 5.25 * phi**2)
            + b1 * (2 + 4.5 * phi + 3 * phi**2)
        )
        m6 = (
            (39 + 94.5 * phi + 52.5 * phi**2)
            + a1 * (21 + 54 * phi + 31.5 * phi**2)
            + b1 * (12.5 + 34.5 * phi + 21 * phi**2)
        )
        m7 = (
            (9 + 21 * phi + 10.5 * phi**2)
            + a1 * (4.5 + 10.5 * phi + 5.25 * phi**2)
            + b1 * (2.5 + 6 * phi + 3 * phi**2)
        )
        m8 = (
            (468 + 882 * phi + 420 * phi**2)
            + a1 * (360 + 672 * phi + 315 * phi**2)
            + b1 * (290 + 540 * phi + 252 * phi**2)
        )
        m9 = (
            (66 + 115.5 * phi + 52.5 * phi**2)
            + a1 * (45 + 75 * phi + 31.5 * phi**2)
            + b1 * (32.5 + 52.5 * phi + 21 * phi**2)
        )
        m10 = (
            (12 + 21 * phi + 10.5 * phi**2)
            + a1 * (7.5 + 12 * phi + 5.25 * phi**2)
            + b1 * (5 + 7.5 * phi + 3 * phi**2)
        )

        # fmt: off
        Mt = np.array([
                [   m1,     0,        0,    L*m2,     m3,     0,        0,    -L*m4],
                [    0,    m1,    -L*m2,       0,      0,    m3,     L*m4,        0],
                [    0, -L*m2,  L**2*m5,       0,      0, -L*m6, -L**2*m7,        0],
                [ L*m2,     0,        0,  L**2*m5,  L*m6,     0,        0, -L**2*m7],
                [   m3,     0,        0,     L*m6,    m8,     0,        0,    -L*m9],
                [    0,    m3,    -L*m6,        0,     0,    m8,     L*m9,        0],
                [    0,  L*m4, -L**2*m7,        0,     0,  L*m9, L**2*m10,        0],
                [-L*m4,     0,        0, -L**2*m7, -L*m9,     0,        0, L**2*m10],
        ])
        # fmt: on
        M = self.material.rho * A_l * L * Mt / (1260 * (1 + phi) ** 2)

        if self.rotary_inertia:
            # fmt: off
            m11 = 252 + 126 * a2 + 72 * b2 + 45 * gama + 30 * delta
            m12 = (
                21 - 105 * phi
                + a2 * (21 - 42 * phi)
                + b2 * (15 - 21 * phi)
                + gama * (10.5 - 12 * phi)
                + delta * (7.5 - 7.5 * phi)
            )
            m13 = (
                21 - 105 * phi
                - 63 * a2 * phi
                - b2 * (6 + 42 * phi)
                - gama * (7.5 + 30 * phi)
                - delta * (7.5 + 22.5 * phi)
            )
            m14 = (
                28 + 35 * phi + 70 * phi ** 2
                + a2 * (7 - 7 * phi + 17.5 * phi ** 2)
                + b2 * (4 - 7 * phi + 7 * phi ** 2)
                + gama * (2.75 - 5 * phi + 3.5 * phi ** 2)
                + delta * (2 - 3.5 * phi + 2 * phi ** 2)
            )
            m15 = (
                7 + 35 * phi - 35 * phi ** 2
                + a2 * (3.5 + 17.5 * phi - 17.5 * phi ** 2)
                + b2 * (3 + 10.5 * phi - 10.5 * phi ** 2)
                + gama * (2.75 + 7 * phi - 7 * phi ** 2)
                + delta * (2.5 + 5 * phi - 5 * phi ** 2)
            )
            m16 = (
                28 + 35 * phi + 70 * phi ** 2
                + a2 * (21 + 42 * phi + 52.5 * phi ** 2)
                + b2 * (18 + 42 * phi + 42 * phi ** 2)
                + gama * (16.25 + 40 * phi + 35 * phi ** 2)
                + delta * (15 + 37.5 * phi + 30 * phi ** 2)
            )

            Mr = np.array([
                    [  m11,      0,         0,     L*m12,   -m11,     0,         0,     L*m13],
                    [    0,    m11,    -L*m12,         0,      0,  -m11,    -L*m13,         0],
                    [    0, -L*m12,  L**2*m14,         0,      0, L*m12, -L**2*m15,         0],
                    [L*m12,      0,         0,  L**2*m14, -L*m12,     0,         0, -L**2*m15],
                    [ -m11,      0,         0,    -L*m12,    m11,     0,         0,    -L*m13],
                    [    0,   -m11,     L*m12,         0,      0,   m11,     L*m13,         0],
                    [    0, -L*m13, -L**2*m15,         0,      0, L*m13,  L**2*m16,         0],
                    [L*m13,      0,         0, -L**2*m15, -L*m13,     0,         0,  L**2*m16],
            ])
            # fmt: on
            Mr = self.material.rho * Ie_l * Mr / (210 * L * (1 + phi) ** 2)
            M = M + Mr

        return M

    def K(self):
        """Stiffness matrix for an instance of a shaft element.

        Returns
        -------
        K : np.ndarray
            Stiffness matrix for the shaft element.

        Examples
        --------
        >>> from ross.materials import steel
        >>> Timoshenko_Element = ShaftElement(
        ...                         L=0.5, idl=0.05, idr=0.05, odl=0.1,
        ...                         odr=0.15, material=steel,
        ...                         rotary_inertia=True,
        ...                         shear_effects=True)
        >>> Timoshenko_Element.K()[:4, :4]/1e6
        array([[219.96144416,   0.        ,   0.        ,  41.29754659],
               [  0.        , 219.96144416, -41.29754659,   0.        ],
               [  0.        , -41.29754659,  12.23526375,   0.        ],
               [ 41.29754659,   0.        ,   0.        ,  12.23526375]])
        """
        L = self.L
        phi = self.phi
        a1 = self.a1
        a2 = self.a2
        b1 = self.b1
        b2 = self.b2
        delta = self.delta
        gama = self.gama
        A = self.A
        A_l = self.A_l
        Ie_l = self.Ie_l
        E = self.material.E

        # fmt: off
        k1 = 1260 + 630 * a2 + 504 * b2 + 441 * gama + 396 * delta
        k2 = (
            630
            + 210 * a2
            + 147 * b2
            + 126 * gama
            + 114 * delta
            - phi * (105 * a2 + 105 * b2 + 94.5 * gama + 84 * delta)
        )
        k3 = (
            630
            + 420 * a2
            + 357 * b2
            + 315 * gama
            + 282 * delta
            + phi * (105 * a2 + 105 * b2 + 94.5 * gama + 84 * delta)
        )
        k4 = (
            420 + 210 * phi + 105 * phi ** 2
            + a2 * (105 + 52.5 * phi ** 2)
            + b2 * (56 - 35 * phi + 35 * phi ** 2)
            + gama * (42 - 42 * phi + 26.25 * phi ** 2)
            + delta * (36 - 42 * phi + 21 * phi ** 2)
        )
        k5 = (
            210 - 210 * phi - 105 * phi ** 2
            + a2 * (105 - 105 * phi - 52.5 * phi ** 2)
            + b2 * (91 - 70 * phi - 35 * phi ** 2)
            + gama * (84 - 52.5 * phi - 26.25 * phi ** 2)
            + delta * (78 - 42 * phi - 21 * phi ** 2)
        )
        k6 = (
            420 + 210 * phi + 105 * phi ** 2
            + a2 * (315 + 210 * phi + 52.5 * phi ** 2)
            + b2 * (266 + 175 * phi + 35 * phi ** 2)
            + gama * (231 + 147 * phi + 26.25 * phi ** 2)
            + delta * (204 + 126 * phi + 21 * phi ** 2)
        )
        k7 = 12 + 6 * a1 + 4 * b1
        k8 = 6 + 3 * a1 + 2 * b1
        k9 = 3 + 1.5 * a1 + b1

        K1 = np.array([
            [  k1,     0,       0,    L*k2,   -k1,    0,       0,    L*k3],
            [   0,    k1,   -L*k2,       0,     0,  -k1,   -L*k3,       0],
            [   0, -L*k2, L**2*k4,       0,     0, L*k2, L**2*k5,       0],
            [L*k2,     0,       0, L**2*k4, -L*k2,    0,       0, L**2*k5],
            [ -k1,     0,       0,   -L*k2,    k1,    0,       0,   -L*k3],
            [   0,   -k1,    L*k2,       0,     0,   k1,    L*k3,       0],
            [   0, -L*k3, L**2*k5,       0,     0, L*k3, L**2*k6,       0],
            [L*k3,     0,       0, L**2*k5, -L*k3,    0,       0, L**2*k6],
        ])

        K2 = np.array([
            [  k7,     0,       0,    L*k8,   -k7,     0,       0,    L*k8],
            [   0,    k7,   -L*k8,       0,     0,   -k7,   -L*k8,       0],
            [   0, -L*k8, L**2*k9,       0,     0,  L*k8, L**2*k9,       0],
            [L*k8,     0,       0, L**2*k9, -L*k8,     0,       0, L**2*k9],
            [ -k7,     0,       0,   -L*k8,    k7,     0,       0,   -L*k8],
            [   0,   -k7,    L*k8,       0,     0,    k7,    L*k8,       0],
            [   0, -L*k8, L**2*k9,       0,     0,  L*k8, L**2*k9,       0],
            [L*k8,     0,       0, L**2*k9, -L*k8,     0,       0, L**2*k9],
        ])

        K = E * L**(-3) * (1 + phi)**(-2) * (K1 * Ie_l/105 + K2 * self.Ie * phi * A_l / A)

        # axial force
        k10 = 36 + 60 * phi + 30 * phi ** 2
        k11 = L * 3
        k12 = L ** 2 * (4 + 5 * phi + 2.5 * phi ** 2)
        k13 = L ** 2 * (1 + 5 * phi + 2.5 * phi ** 2)

        Kaxial = np.array([
            [ k10,    0,    0,  k11, -k10,    0,    0,  k11],
            [   0,  k10, -k11,    0,    0, -k10, -k11,    0],
            [   0, -k11,  k12,    0,    0,  k11, -k13,    0],
            [ k11,    0,    0,  k12, -k11,    0,    0, -k13],
            [-k10,    0,    0, -k11,  k10,    0,    0, -k11],
            [   0, -k10,  k11,    0,    0,  k10,  k11,    0],
            [   0, -k11, -k13,    0,    0,  k11,  k12,    0],
            [ k11,    0,    0, -k13, -k11,    0,    0,  k12],
        ])

        Kaxial *= self.axial_force / (30 * L * (1 + phi) ** 2)
        K += Kaxial

        # torque
        Ktorque = np.array([
            [   0,    0,    1,    0,    0,    0,   -1,    0],
            [   0,    0,    0,    1,    0,    0,    0,   -1],
            [   1,    0,    0, -L/2,   -1,    0,    0,  L/2],
            [   0,    1,  L/2,    0,    0,   -1, -L/2,    0],
            [   0,    0,   -1,    0,    0,    0,    1,    0],
            [   0,    0,    0,   -1,    0,    0,    0,    1],
            [  -1,    0,    0, -L/2,    1,    0,    0,  L/2],
            [   0,   -1,  L/2,    0,    0,    1, -L/2,    0],
        ])

        Ktorque *= self.torque / L
        K += Ktorque
        # fmt: on

        return K

    def C(self):
        """Stiffness matrix for an instance of a shaft element.

        Returns
        -------
        C : np.array
           Damping matrix for the shaft element.

        Examples
        --------
        >>> from ross.materials import steel
        >>> Timoshenko_Element = ShaftElement(
        ...                         L=0.5, idl=0.05, idr=0.05, odl=0.1,
        ...                         odr=0.15, material=steel,
        ...                         rotary_inertia=True,
        ...                         shear_effects=True)
        >>> Timoshenko_Element.C()[:4, :4]
        array([[ 0.,  0.,  0.,  0.],
               [ 0.,  0., -0.,  0.],
               [ 0., -0.,  0.,  0.],
               [ 0.,  0.,  0.,  0.]])
        """
        # proportional damping matrix
        C = self.alpha * self.M() + self.beta * self.K()

        return C

    def G(self):
        """Gyroscopic matrix for an instance of a shaft element.

        Returns
        -------
        G : np.ndarray
            Gyroscopic matrix for the shaft element.

        Examples
        --------
        >>> from ross.materials import steel
        >>> # Timoshenko is the default shaft element
        >>> Timoshenko_Element = ShaftElement(
        ...                         L=0.5, idl=0.05, idr=0.05, odl=0.1,
        ...                         odr=0.15, material=steel,
        ...                         rotary_inertia=True,
        ...                         shear_effects=True)
        >>> Timoshenko_Element.G()[:4, :4]
        array([[ 0.        ,  0.30940809, -0.01085902,  0.        ],
               [-0.30940809,  0.        ,  0.        , -0.01085902],
               [ 0.01085902,  0.        ,  0.        ,  0.0067206 ],
               [ 0.        ,  0.01085902, -0.0067206 ,  0.        ]])
        """
        if self.gyroscopic:
            phi = self.phi
            L = self.L
            a2 = self.a2
            b2 = self.b2
            delta = self.delta
            gama = self.gama
            Ie_l = self.Ie_l

            # fmt: off
            g1 = 252 + 126 * a2 + 72 * b2 + 45 * gama + 30 * delta
            g2 = (
                21 - 105 * phi
                + a2 * (21 - 42 * phi)
                + b2 * (15 - 21 * phi)
                + gama * (10.5 - 12 * phi)
                + delta * (7.5 - 7.5 * phi)
            )
            g3 = (
                21 - 105 * phi
                - 63 * a2 * phi
                - b2 * (6 + 42 * phi)
                - gama * (7.5 + 30 * phi)
                - delta * (7.5 + 22.5 * phi)
            )
            g4 = (
                28 + 35 * phi + 70 * phi ** 2
                + a2 * (7 - 7 * phi + 17.5 * phi ** 2)
                + b2 * (4 - 7 * phi + 7 * phi ** 2)
                + gama * (2.75 - 5 * phi + 3.5 * phi ** 2)
                + delta * (2 - 3.5 * phi + 2 * phi ** 2)
            )
            g5 = (
                7 + 35 * phi - 35 * phi ** 2
                + a2 * (3.5 + 17.5 * phi - 17.5 * phi ** 2)
                + b2 * (3 + 10.5 * phi - 10.5 * phi ** 2)
                + gama * (2.75 + 7 * phi - 7 * phi ** 2)
                + delta * (2.5 + 5 * phi - 5 * phi ** 2)
            )
            g6 = (
                28 + 35 * phi + 70 * phi ** 2
                + a2 * (21 + 42 * phi + 52.5 * phi ** 2)
                + b2 * (18 + 42 * phi + 42 * phi ** 2)
                + gama * (16.25 + 40 * phi + 35 * phi ** 2)
                + delta * (15 + 37.5 * phi + 30 * phi ** 2)
            )

            G = np.array([
                    [   0,    g1,    -L*g2,        0,     0,   -g1,    -L*g3,        0],
                    [ -g1,     0,        0,    -L*g2,    g1,     0,        0,    -L*g3],
                    [L*g2,     0,        0,  L**2*g4, -L*g2,     0,        0, -L**2*g5],
                    [   0,  L*g2, -L**2*g4,        0,     0, -L*g2,  L**2*g5,        0],
                    [   0,   -g1,     L*g2,        0,     0,    g1,     L*g3,        0],
                    [  g1,     0,        0,     L*g2,   -g1,     0,        0,     L*g3],
                    [L*g3,     0,        0, -L**2*g5, -L*g3,     0,        0,  L**2*g6],
                    [   0,  L*g3,  L**2*g5,        0,     0, -L*g3, -L**2*g6,        0],
            ])
            # fmt: on
            G = self.material.rho * Ie_l * 2 * G / (210 * L * (1 + phi) ** 2)

        else:
            G = np.zeros((8, 8))

        return G

    def _patch(self, position, check_sld, fig, units):
        """Shaft element patch.

        Patch that will be used to draw the shaft element using Plotly library.

        Parameters
        ----------
        position : float
            Position in which the patch will be drawn.
        check_sld : bool
            If True, HoverTool displays only the slenderness ratio and color
            the elements in yellow if slenderness ratio < 1.6
        fig : plotly.graph_objects.Figure
            The figure object which traces are added on.
        units : str, optional
            Element length and radius units.
            Default is 'm'.

        Returns
        -------
        fig : plotly.graph_objects.Figure
            The figure object which traces are added on.
        """
        if check_sld is True and self.slenderness_ratio < 1.6:
            color = "yellow"
            legend = "Shaft - Slenderness Ratio < 1.6"
        else:
            color = self.material.color
            legend = "Shaft"

        # plot the shaft
        z_upper = [position, position, position + self.L, position + self.L, position]
        y_upper = [self.idl / 2, self.odl / 2, self.odr / 2, self.idr / 2, self.idl / 2]
        z_lower = [position, position, position + self.L, position + self.L, position]
        y_lower = [
            -self.idl / 2,
            -self.odl / 2,
            -self.odr / 2,
            -self.idr / 2,
            -self.idl / 2,
        ]

        z_pos = z_upper
        z_pos.extend(z_lower)

        y_pos = y_upper
        y_pos.extend(y_lower)

        if check_sld:
            customdata = [self.n, self.slenderness_ratio]
            hovertemplate = (
                f"Element Number: {customdata[0]}<br>"
                + f"Slenderness Ratio: {customdata[1]:.3f}"
            )
        else:
            if isinstance(self, ShaftElement6DoF):
                customdata = [
                    self.n,
                    Q_(self.odl, "m").to(units).m,
                    Q_(self.idl, "m").to(units).m,
                    Q_(self.odr, "m").to(units).m,
                    Q_(self.idr, "m").to(units).m,
                    self.alpha,
                    self.beta,
                    Q_(self.L, "m").to(units).m,
                    self.material.name,
                ]
                hovertemplate = (
                    f"Element Number: {customdata[0]}<br>"
                    + f"Left Outer Diameter: {round(customdata[1], 6)} {units}<br>"
                    + f"Left Inner Diameter: {round(customdata[2], 6)} {units}<br>"
                    + f"Right Outer Diameter: {round(customdata[3], 6)} {units}<br>"
                    + f"Right Inner Diameter: {round(customdata[4], 6)} {units}<br>"
                    + f"Alpha Damp. Factor: {round(customdata[5], 6)}<br>"
                    + f"Beta Damp. Factor: {round(customdata[6], 6)}<br>"
                    + f"Element Length: {round(customdata[7], 6)} {units} <br>"
                    + f"Material: {customdata[8]}<br>"
                )
            else:
                customdata = [
                    self.n,
                    Q_(self.odl, "m").to(units).m,
                    Q_(self.idl, "m").to(units).m,
                    Q_(self.odr, "m").to(units).m,
                    Q_(self.idr, "m").to(units).m,
                    Q_(self.L, "m").to(units).m,
                    self.material.name,
                ]
                hovertemplate = (
                    f"Element Number: {customdata[0]}<br>"
                    + f"Left Outer Diameter: {round(customdata[1], 6)} {units}<br>"
                    + f"Left Inner Diameter: {round(customdata[2], 6)} {units}<br>"
                    + f"Right Outer Diameter: {round(customdata[3], 6)} {units}<br>"
                    + f"Right Inner Diameter: {round(customdata[4], 6)} {units}<br>"
                    + f"Element Length: {round(customdata[5], 6)} {units}<br>"
                    + f"Material: {customdata[6]}<br>"
                )
        fig.add_trace(
            go.Scatter(
                x=Q_(z_pos, "m").to(units).m,
                y=Q_(y_pos, "m").to(units).m,
                customdata=[customdata] * len(z_pos),
                text=hovertemplate,
                mode="lines",
                opacity=0.5,
                fill="toself",
                fillcolor=color,
                line=dict(width=1.5, color="black"),
                showlegend=False,
                name=self.tag,
                legendgroup=legend,
                hoveron="points+fills",
                hoverinfo="text",
                hovertemplate=hovertemplate,
                hoverlabel=dict(bgcolor=color),
            )
        )

        return fig

    def create_modified(self, **attributes):
        """Return a new shaft element based on the current instance.

        Any attribute passed as an argument will be used to modify the corresponding
        attribute of the instance. Attributes not provided as arguments will retain
        their values from the current instance.

        Parameters
        ----------
        L : float, pint.Quantity, optional
            Element length (m). Default is equal to value of current instance.
        idl : float, pint.Quantity, optional
            Inner diameter of the element at the left position (m).
            Default is equal to value of current instance.
        odl : float, pint.Quantity, optional
            Outer diameter of the element at the left position (m).
            Default is equal to value of current instance.
        idr : float, pint.Quantity, optional
            Inner diameter of the element at the right position (m).
            Default is equal to value of current instance.
        odr : float, pint.Quantity, optional
            Outer diameter of the element at the right position (m).
            Default is equal to value of current instance.
        material : ross.Material, optional
            Shaft material. Default is equal to value of current instance.
        n : int, optional
            Element number (coincident with it's first node).
            Default is equal to value of current instance.
        axial_force : float, optional
            Axial force (N). Default is equal to value of current instance.
        torque : float, optional
            Torque (N*m). Default is equal to value of current instance.
        shear_effects : bool, optional
            Determine if shear effects are taken into account.
            Default is equal to value of current instance.
        rotary_inertia : bool, optional
            Determine if rotary_inertia effects are taken into account.
            Default is equal to value of current instance.
        gyroscopic : bool, optional
            Determine if gyroscopic effects are taken into account.
            Default is equal to value of current instance.
        shear_method_calc : str, optional
            Determines which shear calculation method the user will adopt
            Default is equal to value of current instance.
        alpha : float, optional
            Mass proportional damping factor.
            Default is equal to value of current instance.
        beta : float, optional
            Stiffness proportional damping factor.
            Default is equal to value of current instance.
        tag : str, optional
            Element tag.
            Default is None.

        Returns
        -------
        shaft_element : ross.ShaftElement
            An instance of the modified shaft element.
        """
        return self.__class__(
            L=attributes.get("L", self.L),
            idl=attributes.get("idl", self.idl),
            odl=attributes.get("odl", self.odl),
            idr=attributes.get("idr", self.idr),
            odr=attributes.get("odr", self.odr),
            n=attributes.get("n", self.n),
            material=attributes.get("material", self.material),
            axial_force=attributes.get("axial_force", self.axial_force),
            torque=attributes.get("torque", self.torque),
            rotary_inertia=attributes.get("rotary_inertia", self.rotary_inertia),
            shear_effects=attributes.get("shear_effects", self.shear_effects),
            gyroscopic=attributes.get("gyroscopic", self.gyroscopic),
            shear_method_calc=attributes.get(
                "shear_method_calc", self.shear_method_calc
            ),
            alpha=attributes.get("alpha", self.alpha),
            beta=attributes.get("beta", self.beta),
            tag=attributes.get("tag", None),
        )

    @classmethod
    def from_table(cls, file, sheet_type="Simple", sheet_name=0):
        """Instantiate one or more shafts using inputs from an Excel table.

        A header with the names of the columns is required. These names should
        match the names expected by the routine (usually the names of the
        parameters, but also similar ones). The program will read every row
        bellow the header until they end or it reaches a NaN.

        Parameters
        ----------
        file : str
            Path to the file containing the shaft parameters.
        sheet_type : str, optional
            Describes the kind of sheet the function should expect:
                Simple: The input table should specify only the number of the materials
                to be used.
                They must be saved prior to calling the method.
                Model: The materials parameters must be passed along with the shaft
                parameters. Each material must have an id number and each shaft must
                reference one of the materials ids.
        sheet_name : int or str, optional
            Position of the sheet in the file (starting from 0) or its name. If none is
            passed, it is assumed to be the first sheet in the file.

        Returns
        -------
        shaft : list
            A list of shaft objects.
        """
        parameters = read_table_file(
            file, "shaft", sheet_name=sheet_name, sheet_type=sheet_type
        )
        list_of_shafts = []
        if sheet_type == "Model":
            new_materials = {}
            for i in range(0, len(parameters["matno"])):
                new_material = Material(
                    name="shaft_mat_" + str(parameters["matno"][i]),
                    rho=parameters["rhoa"][i],
                    E=parameters["ea"][i],
                    G_s=parameters["ga"][i],
                    color=parameters["color"][i],
                )
                new_materials["shaft_mat_" + str(parameters["matno"][i])] = new_material
            for i in range(0, len(parameters["L"])):
                list_of_shafts.append(
                    cls(
                        L=parameters["L"][i],
                        idl=parameters["idl"][i],
                        odl=parameters["odl"][i],
                        idr=parameters["idr"][i],
                        odr=parameters["odr"][i],
                        material=new_materials[parameters["material"][i]],
                        n=parameters["n"][i],
                        axial_force=parameters["axial_force"][i],
                        torque=parameters["torque"][i],
                        shear_effects=parameters["shear_effects"][i],
                        rotary_inertia=parameters["rotary_inertia"][i],
                        gyroscopic=parameters["gyroscopic"][i],
                        shear_method_calc=parameters["shear_method_calc"][i],
                        alpha=parameters["alpha"][i] if "alpha" in parameters else 0,
                        beta=parameters["beta"][i] if "beta" in parameters else 0,
                    )
                )
        elif sheet_type == "Simple":
            for i in range(0, len(parameters["L"])):
                list_of_shafts.append(
                    cls(
                        L=parameters["L"][i],
                        idl=parameters["idl"][i],
                        odl=parameters["odl"][i],
                        idr=parameters["idr"][i],
                        odr=parameters["odr"][i],
                        material=parameters["material"][i],
                        n=parameters["n"][i],
                        axial_force=parameters["axial_force"][i],
                        torque=parameters["torque"][i],
                        shear_effects=parameters["shear_effects"][i],
                        rotary_inertia=parameters["rotary_inertia"][i],
                        gyroscopic=parameters["gyroscopic"][i],
                        shear_method_calc=parameters["shear_method_calc"][i],
                        alpha=parameters["alpha"][i] if "alpha" in parameters else 0,
                        beta=parameters["beta"][i] if "beta" in parameters else 0,
                    )
                )
        return list_of_shafts

    @classmethod
    def section(
        cls,
        L,
        ne,
        s_idl,
        s_odl,
        s_idr=None,
        s_odr=None,
        material=None,
        n=None,
        shear_effects=True,
        rotary_inertia=True,
        gyroscopic=True,
        alpha=0,
        beta=0,
    ):
        """Shaft section constructor.

        This method will create a shaft section with length 'L' divided into
        'ne' elements.

        Parameters
        ----------
        i_d : float
            Inner diameter of the section.
        o_d : float
            Outer diameter of the section.
        E : float
            Young's modulus.
        G_s : float
            Shear modulus.
        material : ross.material
            Shaft material.
        n : int, optional
            Element number (coincident with it's first node).
            If not given, it will be set when the rotor is assembled
            according to the element's position in the list supplied to
            the rotor constructor.
        axial_force : float
            Axial force.
        torque : float
            Torque.
        shear_effects : bool
            Determine if shear effects are taken into account.
            Default is False.
        rotary_inertia : bool
            Determine if rotary_inertia effects are taken into account.
            Default is False.
        gyroscopic : bool
            Determine if gyroscopic effects are taken into account.
            Default is False.
        alpha : float, optional
            Mass proportional damping factor.
            Default is zero.
        beta : float, optional
            Stiffness proportional damping factor.
            Default is zero.

        Returns
        -------
        elements : list
            List with the 'ne' shaft elements.

        Examples
        --------
        >>> # shaft material
        >>> from ross.materials import steel
        >>> # shaft inner and outer diameters
        >>> s_idl = 0
        >>> s_odl = 0.01585
        >>> sec = ShaftElement.section(247.65e-3, 4, 0, 15.8e-3, material=steel)
        >>> len(sec)
        4
        >>> sec[0].i_d
        0.0
        """
        if s_idr is None:
            s_idr = s_idl
        if s_odr is None:
            s_odr = s_odl

        le = L / ne

        elements = [
            cls(
                L=le,
                idl=(s_idr - s_idl) * i * le / L + s_idl,
                odl=(s_odr - s_odl) * i * le / L + s_odl,
                idr=(s_idr - s_idl) * (i + 1) * le / L + s_idl,
                odr=(s_odr - s_odl) * (i + 1) * le / L + s_odl,
                material=material,
                n=n,
                shear_effects=shear_effects,
                rotary_inertia=rotary_inertia,
                gyroscopic=gyroscopic,
                alpha=alpha,
                beta=beta,
            )
            for i in range(ne)
        ]

        return elements


class ShaftElement6DoF(ShaftElement):
    r"""A shaft element with 6 degrees of freedom.

    This class will create a shaft element that may take into
    account shear, rotary inertia an gyroscopic effects. The object can be
    cylindrical or conical and the formulation is based on :cite:`genta1988conical`
    The matrices will be defined considering the following local
    coordinate vector:

    .. math::

        [x_0, y_0, z_0, \alpha_0, \beta_0, \theta_0, x_1, y_1, z_1, \alpha_1, \beta_1, \theta_1]^T
    Where :math:`\x_0` and :math:`\x_1` correspond to horizontal translation,
    :math:`\y_0` and :math:`\y_1` correspond to vertical translation,
    :math:`\z_0` and :math:`\z_1` correspond to axial translation,
    :math:`\alpha_0` and :math:`\alpha_1` correspond to bending on the yz plane,
    :math:`\beta_0` and :math:`\beta_1` correspond to bending on the xz plane
    and :math:`\theta_0` and :math:`\theta_1` correspond to torsion around z axis.

    Parameters
    ----------
    L : float, pint.Quantity
        Element length (m).
    idl : float, pint.Quantity
        Inner diameter of the element at the left position (m).
    odl : float, pint.Quantity
        Outer diameter of the element at the left position (m).
    idr : float, pint.Quantity, optional
        Inner diameter of the element at the right position (m).
        Default is equal to idl value (cylindrical element).
    odr : float, pint.Quantity, optional
        Outer diameter of the element at the right position (m).
        Default is equal to odl value (cylindrical element).
    material : ross.Material
        Shaft material.
    n : int, optional
        Element number (coincident with it's first node).
        If not given, it will be set when the rotor is assembled
        according to the element's position in the list supplied to
        the rotor constructor.
    axial_force : float, optional
        Axial force (N).
    torque : float, optional
        Torque (N*m).
    shear_effects : bool, optional
        Determine if shear effects are taken into account.
        Default is True.
    rotary_inertia : bool, optional
        Determine if rotary_inertia effects are taken into account.
        Default is True.
    gyroscopic : bool, optional
        Determine if gyroscopic effects are taken into account.
        Default is True.
    shear_method_calc : str, optional
        Determines which shear calculation method the user will adopt
        Default is 'cowper'
    alpha : float, optional
        Mass proportional damping factor.
        Default is zero.
    beta : float, optional
        Stiffness proportional damping factor.
        Default is zero.
    tag : str, optional
        Element tag.
        Default is None.

    Returns
    -------
    shaft_element : ross.ShaftElement6DoF
        A 6 degrees of freedom shaft element object.

    Attributes
    ----------
    Poisson : float
        Poisson coefficient for the element.
    A : float
        Element section area at half length (m**2).
    A_l : float
        Element section area at left end (m**2).
    A_r : float
        Element section area at right end (m**2).
    beam_cg : float
        Element center of gravity local position (m).
    axial_cg_pos : float
        Element center of gravity global position (m).
        This should be used only after the rotor is built.
        Default is None.
    Ie : float
        Ie is the second moment of area of the cross section about
        the neutral plane (m**4).
    phi : float
        Constant that is used according to :cite:`friswell2010dynamics` to
        consider rotary inertia and shear effects. If these are not considered
        :math:`\phi=0`.
    kappa : float
        Shear coefficient for the element.

    References
    ----------
    .. bibliography::
        :filter: docname in docnames

    Examples
    --------
    >>> from ross.materials import steel
    >>> # Euler-Bernoulli conical element
    >>> Euler_Bernoulli_Element = ShaftElement6DoF(
    ...                         material=steel, L=0.5, idl=0.05, odl=0.1,
    ...                         idr=0.05, odr=0.15,
    ...                         rotary_inertia=False,
    ...                         shear_effects=False)
    >>> Euler_Bernoulli_Element.phi
    0
    >>> # Timoshenko cylindrical element. In this case idr and odr are omitted.
    >>> Timoshenko_Element = ShaftElement6DoF(
    ...                         material=steel, L=0.5, idl=0.05, odl=0.1,
    ...                         rotary_inertia=True,
    ...                         shear_effects=True)
    >>> Timoshenko_Element.phi
    0.1571268472906404
    """

    def dof_mapping(self):
        """Degrees of freedom mapping.

        Returns a dictionary with a mapping between degree of freedom and its index.

        Returns
        -------
        dof_mapping : dict
            A dictionary containing the degrees of freedom and their indexes
            with the following ordering:
            x_0 - horizontal translation
            y_0 - vertical translation
            z_0 - axial translation
            alpha_0 - rotation around horizontal
            beta_0  - rotation around vertical
            theta_0 - torsion around axial

        Examples
        --------
        >>> sh = ShaftElement6DoF(L=0.5, idl=0.05, odl=0.1, material=steel,
        ...                       rotary_inertia=True, shear_effects=True)
        >>> sh.dof_mapping()["x_0"]
        0
        """
        return dict(
            x_0=0,
            y_0=1,
            z_0=2,
            alpha_0=3,
            beta_0=4,
            theta_0=5,
            x_1=6,
            y_1=7,
            z_1=8,
            alpha_1=9,
            beta_1=10,
            theta_1=11,
        )

    def M(self):
        """Mass matrix for an instance of a 6 DoF shaft element.

        Returns
        -------
        M : np.ndarray
            Mass matrix for the 6 DoF shaft element.

        Examples
        --------
        >>> Timoshenko_Element = ShaftElement6DoF(0.25, 0, 0.05, material=steel)
        >>> Timoshenko_Element.M().shape
        (12, 12)
        """
        phi = self.phi
        L = self.L
        a1 = self.a1
        a2 = self.a2
        b1 = self.b1
        b2 = self.b2
        delta = self.delta
        gama = self.gama

        m1 = (
            (468 + 882 * phi + 420 * phi**2)
            + a1 * (108 + 210 * phi + 105 * phi**2)
            + b1 * (38 + 78 * phi + 42 * phi**2)
        )
        m2 = (
            (66 + 115.5 * phi + 52.5 * phi**2)
            + a1 * (21 + 40.5 * phi + 21 * phi**2)
            + b1 * (8.5 + 18 * phi + 10.5 * phi**2)
        ) * L
        m3 = (
            (162 + 378 * phi + 210 * phi**2)
            + a1 * (81 + 189 * phi + 105 * phi**2)
            + b1 * (46 + 111 * phi + 63 * phi**2)
        )
        m4 = (
            (39 + 94.5 * phi + 52.5 * phi**2)
            + a1 * (18 + 40.5 * phi + 21 * phi**2)
            + b1 * (9.5 + 21 * phi + 10.5 * phi**2)
        ) * L
        m5 = (
            (12 + 21 * phi + 10.5 * phi**2)
            + a1 * (4.5 + 9 * phi + 5.25 * phi**2)
            + b1 * (2 + 4.5 * phi + 3 * phi**2)
        ) * L**2
        m6 = (
            (39 + 94.5 * phi + 52.5 * phi**2)
            + a1 * (21 + 54 * phi + 31.5 * phi**2)
            + b1 * (12.5 + 34.5 * phi + 21 * phi**2)
        ) * L
        m7 = (
            (9 + 21 * phi + 10.5 * phi**2)
            + a1 * (4.5 + 10.5 * phi + 5.25 * phi**2)
            + b1 * (2.5 + 6 * phi + 3 * phi**2)
        ) * L**2
        m8 = (
            (468 + 882 * phi + 420 * phi**2)
            + a1 * (360 + 672 * phi + 315 * phi**2)
            + b1 * (290 + 540 * phi + 252 * phi**2)
        )
        m9 = (
            (66 + 115.5 * phi + 52.5 * phi**2)
            + a1 * (45 + 75 * phi + 31.5 * phi**2)
            + b1 * (32.5 + 52.5 * phi + 21 * phi**2)
        ) * L
        m10 = (
            (12 + 21 * phi + 10.5 * phi**2)
            + a1 * (7.5 + 12 * phi + 5.25 * phi**2)
            + b1 * (5 + 7.5 * phi + 3 * phi**2)
        ) * L**2

        # fmt: off
        M = np.array([
            [ m1,   0, 0,   0,  m2, 0,  m3,   0, 0,   0, -m4, 0],
            [  0,  m1, 0, -m2,   0, 0,   0,  m3, 0,  m4,   0, 0],
            [  0,   0, 0,   0,   0, 0,   0,   0, 0,   0,   0, 0],
            [  0, -m2, 0,  m5,   0, 0,   0, -m6, 0, -m7,   0, 0],
            [ m2,   0, 0,   0,  m5, 0,  m6,   0, 0,   0, -m7, 0],
            [  0,   0, 0,   0,   0, 0,   0,   0, 0,   0,   0, 0],
            [ m3,   0, 0,   0,  m6, 0,  m8,   0, 0,   0, -m9, 0],
            [  0,  m3, 0, -m6,   0, 0,   0,  m8, 0,  m9,   0, 0],
            [  0,   0, 0,   0,   0, 0,   0,   0, 0,   0,   0, 0],
            [  0,  m4, 0, -m7,   0, 0,   0,  m9, 0, m10,   0, 0],
            [-m4,   0, 0,   0, -m7, 0, -m9,   0, 0,   0, m10, 0],
            [  0,   0, 0,   0,   0, 0,   0,   0, 0,   0,   0, 0],
        ]) * self.material.rho * self.A_l * L / (1260 * (1 + phi) ** 2)
        # fmt: on

        if self.rotary_inertia:
            # fmt: off
            m11 = 252 + 126 * a2 + 72 * b2 + 45 * gama + 30 * delta
            m12 = (
                21 - 105 * phi
                + a2 * (21 - 42 * phi)
                + b2 * (15 - 21 * phi)
                + gama * (10.5 - 12 * phi)
                + delta * (7.5 - 7.5 * phi)
            ) * L
            m13 = (
                21 - 105 * phi
                - 63 * a2 * phi
                - b2 * (6 + 42 * phi)
                - gama * (7.5 + 30 * phi)
                - delta * (7.5 + 22.5 * phi)
            ) * L
            m14 = (
                28 + 35 * phi + 70 * phi ** 2
                + a2 * (7 - 7 * phi + 17.5 * phi ** 2)
                + b2 * (4 - 7 * phi + 7 * phi ** 2)
                + gama * (2.75 - 5 * phi + 3.5 * phi ** 2)
                + delta * (2 - 3.5 * phi + 2 * phi ** 2)
            ) * L ** 2
            m15 = (
                7 + 35 * phi - 35 * phi ** 2
                + a2 * (3.5 + 17.5 * phi - 17.5 * phi ** 2)
                + b2 * (3 + 10.5 * phi - 10.5 * phi ** 2)
                + gama * (2.75 + 7 * phi - 7 * phi ** 2)
                + delta * (2.5 + 5 * phi - 5 * phi ** 2)
            ) * L ** 2
            m16 = (
                28 + 35 * phi + 70 * phi ** 2
                + a2 * (21 + 42 * phi + 52.5 * phi ** 2)
                + b2 * (18 + 42 * phi + 42 * phi ** 2)
                + gama * (16.25 + 40 * phi + 35 * phi ** 2)
                + delta * (15 + 37.5 * phi + 30 * phi ** 2)
            ) * L ** 2

            Mr = np.array([
                [ m11,    0, 0,    0,  m12, 0, -m11,    0, 0,    0,  m13, 0],
                [   0,  m11, 0, -m12,    0, 0,    0, -m11, 0, -m13,    0, 0],
                [   0,    0, 0,    0,    0, 0,    0,    0, 0,    0,    0, 0],
                [   0, -m12, 0,  m14,    0, 0,    0,  m12, 0, -m15,    0, 0],
                [ m12,    0, 0,    0,  m14, 0, -m12,    0, 0,    0, -m15, 0],
                [   0,    0, 0,    0,    0, 0,    0,    0, 0,    0,    0, 0],
                [-m11,    0, 0,    0, -m12, 0,  m11,    0, 0,    0, -m13, 0],
                [   0, -m11, 0,  m12,    0, 0,    0,  m11, 0,  m13,    0, 0],
                [   0,    0, 0,    0,    0, 0,    0,    0, 0,    0,    0, 0],
                [   0, -m13, 0, -m15,    0, 0,    0,  m13, 0,  m16,    0, 0],
                [ m13,    0, 0,    0, -m15, 0, -m13,    0, 0,    0,  m16, 0],
                [   0,    0, 0,    0,    0, 0,    0,    0, 0,    0,    0, 0],
            ]) * self.material.rho * self.Ie_l / (210 * L * (1 + phi) ** 2)
            # fmt: on
            M += Mr

        # axial motion
        Ae = (self.A_l + self.A_r) / 2
        # fmt: off
        Mam = np.array([
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 2, 0, 0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0, 2, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        ]) * self.material.rho * Ae * L / 6
        # fmt: on
        M += Mam

        # torsion
        Je = 2 * (self.Ie_l + self.Ie_r) / 2
        # fmt: off
        Mts = np.array([
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 2],
        ]) * self.material.rho * Je * L / 6
        # fmt: on
        M += Mts

        return M

    def K(self):
        """Stiffness matrix for an instance of a 6 DoF shaft element.

        Returns
        -------
        K : np.ndarray
            Stiffness matrix for the 6 DoF shaft element.

        Examples
        --------
        >>> from ross.materials import steel
        >>> Timoshenko_Element = ShaftElement6DoF(0.25, 0, 0.05, material=steel)
        >>> Timoshenko_Element.K().shape
        (12, 12)
        """
        L = self.L
        phi = self.phi
        a1 = self.a1
        a2 = self.a2
        b1 = self.b1
        b2 = self.b2
        delta = self.delta
        gama = self.gama

        # fmt: off
        k1 = 1260 + 630 * a2 + 504 * b2 + 441 * gama + 396 * delta
        k2 = (
            630
            + 210 * a2
            + 147 * b2
            + 126 * gama
            + 114 * delta
            - phi * (105 * a2 + 105 * b2 + 94.5 * gama + 84 * delta)
        ) * L
        k3 = (
            630
            + 420 * a2
            + 357 * b2
            + 315 * gama
            + 282 * delta
            + phi * (105 * a2 + 105 * b2 + 94.5 * gama + 84 * delta)
        ) * L
        k4 = (
            420 + 210 * phi + 105 * phi ** 2
            + a2 * (105 + 52.5 * phi ** 2)
            + b2 * (56 - 35 * phi + 35 * phi ** 2)
            + gama * (42 - 42 * phi + 26.25 * phi ** 2)
            + delta * (36 - 42 * phi + 21 * phi ** 2)
        ) * L ** 2
        k5 = (
            210 - 210 * phi - 105 * phi ** 2
            + a2 * (105 - 105 * phi - 52.5 * phi ** 2)
            + b2 * (91 - 70 * phi - 35 * phi ** 2)
            + gama * (84 - 52.5 * phi - 26.25 * phi ** 2)
            + delta * (78 - 42 * phi - 21 * phi ** 2)
        ) * L ** 2
        k6 = (
            420 + 210 * phi + 105 * phi ** 2
            + a2 * (315 + 210 * phi + 52.5 * phi ** 2)
            + b2 * (266 + 175 * phi + 35 * phi ** 2)
            + gama * (231 + 147 * phi + 26.25 * phi ** 2)
            + delta * (204 + 126 * phi + 21 * phi ** 2)
        ) * L ** 2
        k7 = 12 + 6 * a1 + 4 * b1
        k8 = (6 + 3 * a1 + 2 * b1) * L
        k9 = (3 + 1.5 * a1 + b1) * L ** 2

        K1 = np.array([
            [ k1,   0,  0,   0,  k2,  0, -k1,   0,  0,   0,  k3,  0],
            [  0,  k1,  0, -k2,   0,  0,   0, -k1,  0, -k3,   0,  0],
            [  0,   0,  0,   0,   0,  0,   0,   0,  0,   0,   0,  0],
            [  0, -k2,  0,  k4,   0,  0,   0,  k2,  0,  k5,   0,  0],
            [ k2,   0,  0,   0,  k4,  0, -k2,   0,  0,   0,  k5,  0],
            [  0,   0,  0,   0,   0,  0,   0,   0,  0,   0,   0,  0],
            [-k1,   0,  0,   0, -k2,  0,  k1,   0,  0,   0, -k3,  0],
            [  0, -k1,  0,  k2,   0,  0,   0,  k1,  0,  k3,   0,  0],
            [  0,   0,  0,   0,   0,  0,   0,   0,  0,   0,   0,  0],
            [  0, -k3,  0,  k5,   0,  0,   0,  k3,  0,  k6,   0,  0],
            [ k3,   0,  0,   0,  k5,  0, -k3,   0,  0,   0,  k6,  0],
            [  0,   0,  0,   0,   0,  0,   0,   0,  0,   0,   0,  0],
        ]) * self.Ie_l / 105 

        K2 = np.array([
            [ k7,   0,  0,   0,  k8,  0, -k7,   0,  0,   0,  k8,  0],
            [  0,  k7,  0, -k8,   0,  0,   0, -k7,  0, -k8,   0,  0],
            [  0,   0,  0,   0,   0,  0,   0,   0,  0,   0,   0,  0],
            [  0, -k8,  0,  k9,   0,  0,   0,  k8,  0,  k9,   0,  0],
            [ k8,   0,  0,   0,  k9,  0, -k8,   0,  0,   0,  k9,  0],
            [  0,   0,  0,   0,   0,  0,   0,   0,  0,   0,   0,  0],
            [-k7,   0,  0,   0, -k8,  0,  k7,   0,  0,   0, -k8,  0],
            [  0, -k7,  0,  k8,   0,  0,   0,  k7,  0,  k8,   0,  0],
            [  0,   0,  0,   0,   0,  0,   0,   0,  0,   0,   0,  0],
            [  0, -k8,  0,  k9,   0,  0,   0,  k8,  0,  k9,   0,  0],
            [ k8,   0,  0,   0,  k9,  0, -k8,   0,  0,   0,  k9,  0],
            [  0,   0,  0,   0,   0,  0,   0,   0,  0,   0,   0,  0],
        ]) * self.Ie * phi * self.A_l / self.A
        # fmt: on

        K = self.material.E * L ** (-3) * (1 + phi) ** (-2) * (K1 + K2)

        # axial force
        k10 = 36 + 60 * phi + 30 * phi**2
        k11 = 3 * L
        k12 = (4 + 5 * phi + 2.5 * phi**2) * L**2
        k13 = (1 + 5 * phi + 2.5 * phi**2) * L**2
        # fmt: off
        Kaf = np.array([
            [ k10,    0,  0,    0,  k11,  0, -k10,    0,  0,    0,  k11,  0],
            [   0,  k10,  0, -k11,    0,  0,    0, -k10,  0, -k11,    0,  0],
            [   0,    0,  0,    0,    0,  0,    0,    0,  0,    0,    0,  0],
            [   0, -k11,  0,  k12,    0,  0,    0,  k11,  0, -k13,    0,  0],
            [ k11,    0,  0,    0,  k12,  0, -k11,    0,  0,    0, -k13,  0],
            [   0,    0,  0,    0,    0,  0,    0,    0,  0,    0,    0,  0],
            [-k10,    0,  0,    0, -k11,  0,  k10,    0,  0,    0, -k11,  0],
            [   0, -k10,  0,  k11,    0,  0,    0,  k10,  0,  k11,    0,  0],
            [   0,    0,  0,    0,    0,  0,    0,    0,  0,    0,    0,  0],
            [   0, -k11,  0, -k13,    0,  0,    0,  k11,  0,  k12,    0,  0],
            [ k11,    0,  0,    0, -k13,  0, -k11,    0,  0,    0,  k12,  0],
            [   0,    0,  0,    0,    0,  0,    0,    0,  0,    0,    0,  0],
        ]) * self.axial_force / (30 * L * (1 + phi) ** 2)
        # fmt: on
        K += Kaf

        # torque
        # fmt: off
        Ktq = np.array([
            [ 0,  0,  0,    1,    0,  0,  0,  0,  0,   -1,    0,  0],
            [ 0,  0,  0,    0,    1,  0,  0,  0,  0,    0,   -1,  0],
            [ 0,  0,  0,    0,    0,  0,  0,  0,  0,    0,    0,  0],
            [ 1,  0,  0,    0, -L/2,  0, -1,  0,  0,    0,  L/2,  0],
            [ 0,  1,  0,  L/2,    0,  0,  0, -1,  0, -L/2,    0,  0],
            [ 0,  0,  0,    0,    0,  0,  0,  0,  0,    0,    0,  0],
            [ 0,  0,  0,   -1,    0,  0,  0,  0,  0,    1,    0,  0],
            [ 0,  0,  0,    0,   -1,  0,  0,  0,  0,    0,    1,  0],
            [ 0,  0,  0,    0,    0,  0,  0,  0,  0,    0,    0,  0],
            [-1,  0,  0,    0, -L/2,  0,  1,  0,  0,    0,  L/2,  0],
            [ 0, -1,  0,  L/2,    0,  0,  0,  1,  0, -L/2,    0,  0],
            [ 0,  0,  0,    0,    0,  0,  0,  0,  0,    0,    0,  0],
        ]) * self.torque / L
        # fmt: on
        K += Ktq

        # axial motion
        Ae = (self.A_l + self.A_r) / 2
        # fmt: off
        Kam = np.array([
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  1,  0,  0,  0,  0,  0, -1,  0,  0,  0],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0, -1,  0,  0,  0,  0,  0,  1,  0,  0,  0],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
        ]) * self.material.E * Ae / L
        # fmt: on
        K += Kam

        # torsion
        Je = 2 * (self.Ie_l + self.Ie_r) / 2
        # fmt: off
        Kts = np.array([
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  0,  0,  0,  1,  0,  0,  0,  0,  0, -1],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0],
            [0,  0,  0,  0,  0, -1,  0,  0,  0,  0,  0,  1],
        ]) * self.material.G_s * Je / L
        # fmt: on
        K += Kts

        return K

    def Kst(self):
        """Dynamic stiffness matrix for an instance of a 6 DoF shaft element.

        Stiffness matrix for the 6 DoF shaft element associated with
        the transient motion. It needs to be multiplied by the angular
        acceleration when considered in time dependent analyses.

        Returns
        -------
        Kst : np.ndarray
            Dynamic stiffness matrix for the 6 DoF shaft element.

        Examples
        --------
        >>> from ross.materials import steel
        >>> Timoshenko_Element = ShaftElement6DoF(0.25, 0, 0.05, material=steel)
        >>> Timoshenko_Element.Kst().shape
        (12, 12)
        """
        L = self.L
        Ie = (self.Ie_l + self.Ie_r) / 2

        # fmt: off
        Kst = np.array([
            [0,  -36,  0,    3*L,  0,  0,  0,  36,  0,    3*L,  0,  0],
            [0,    0,  0,      0,  0,  0,  0,   0,  0,      0,  0,  0],
            [0,    0,  0,      0,  0,  0,  0,   0,  0,      0,  0,  0],
            [0,    0,  0,      0,  0,  0,  0,   0,  0,      0,  0,  0],
            [0, -3*L,  0, 4*L**2,  0,  0,  0, 3*L,  0,  -L**2,  0,  0],
            [0,    0,  0,      0,  0,  0,  0,   0,  0,      0,  0,  0],
            [0,   36,  0,   -3*L,  0,  0,  0, -36,  0,   -3*L,  0,  0],
            [0,    0,  0,      0,  0,  0,  0,   0,  0,      0,  0,  0],
            [0,    0,  0,      0,  0,  0,  0,   0,  0,      0,  0,  0],
            [0,    0,  0,      0,  0,  0,  0,   0,  0,      0,  0,  0],
            [0, -3*L,  0,  -L**2,  0,  0,  0, 3*L,  0, 4*L**2,  0,  0],
            [0,    0,  0,      0,  0,  0,  0,   0,  0,      0,  0,  0],
        ]) * self.material.rho * Ie / (15 * L) 
        # fmt: on

        return Kst

    def C(self):
        """Proportional damping matrix for an instance of a 6 DoF shaft element.

        Returns
        -------
        C : np.ndarray
            Proportional damping matrix for the 6 DoF shaft element.

        Examples
        --------
        >>> from ross.materials import steel
        >>> shaft = ShaftElement6DoF(L=0.25, idl=0, odl=0.05, material=steel)
        >>> shaft.C().shape
        (12, 12)
        """
        # proportional damping matrix
        C = self.alpha * self.M() + self.beta * self.K()

        return C

    def G(self):
        """Gyroscopic matrix for an instance of a 6 DoFs shaft element.

        Gyroscopic matrix for the 6 DoF shaft element. It needs to be
        multiplied by the angular velocity when considered in time
        dependent analyses.

        Returns
        -------
        G : np.ndarray
            Gyroscopic matrix for the 6 DoF shaft element.

        Examples
        --------
        >>> from ross.materials import steel
        >>> shaft = ShaftElement6DoF(0.25, 0, 0.05, material=steel)
        >>> shaft.G().shape
        (12, 12)
        """
        if self.gyroscopic:
            phi = self.phi
            L = self.L
            a2 = self.a2
            b2 = self.b2
            delta = self.delta
            gama = self.gama

            # fmt: off
            g1 = 252 + 126 * a2 + 72 * b2 + 45 * gama + 30 * delta
            g2 = (
                21
                - 105 * phi
                + a2 * (21 - 42 * phi)
                + b2 * (15 - 21 * phi)
                + gama * (10.5 - 12 * phi)
                + delta * (7.5 - 7.5 * phi)
            ) * L
            g3 = (
                21
                - 105 * phi
                - 63 * a2 * phi
                - b2 * (6 + 42 * phi)
                - gama * (7.5 + 30 * phi)
                - delta * (7.5 + 22.5 * phi)
            ) * L
            g4 = (
                28
                + 35 * phi
                + 70 * phi ** 2
                + a2 * (7 - 7 * phi + 17.5 * phi ** 2)
                + b2 * (4 - 7 * phi + 7 * phi ** 2)
                + gama * (2.75 - 5 * phi + 3.5 * phi ** 2)
                + delta * (2 - 3.5 * phi + 2 * phi ** 2)
            ) * L ** 2
            g5 = (
                7
                + 35 * phi
                - 35 * phi ** 2
                + a2 * (3.5 + 17.5 * phi - 17.5 * phi ** 2)
                + b2 * (3 + 10.5 * phi - 10.5 * phi ** 2)
                + gama * (2.75 + 7 * phi - 7 * phi * 2)
                + delta * (2.5 + 5 * phi - 5 * phi ** 2)
            ) * L ** 2
            g6 = (
                28
                + 35 * phi
                + 70 * phi**2
                + a2 * (21 + 42 * phi + 52.5 * phi ** 2)
                + b2 * (18 + 42 * phi + 42 * phi ** 2)
                + gama * (16.25 + 40 * phi + 35 * phi ** 2)
                + delta * (15 + 37.5 * phi + 30 * phi ** 2)
            ) * L ** 2

            G = np.array([
                [  0,  g1,  0, -g2,   0,  0,   0, -g1,  0, -g3,   0,  0],
                [-g1,   0,  0,   0, -g2,  0,  g1,   0,  0,   0, -g3,  0],
                [  0,   0,  0,   0,   0,  0,   0,   0,  0,   0,   0,  0],
                [ g2,   0,  0,   0,  g4,  0, -g2,   0,  0,   0, -g5,  0],
                [  0,  g2,  0, -g4,   0,  0,   0, -g2,  0,  g5,   0,  0],
                [  0,   0,  0,   0,   0,  0,   0,   0,  0,   0,   0,  0],
                [  0, -g1,  0,  g2,   0,  0,   0,  g1,  0,  g3,   0,  0],
                [ g1,   0,  0,   0,  g2,  0, -g1,   0,  0,   0,  g3,  0],
                [  0,   0,  0,   0,   0,  0,   0,   0,  0,   0,   0,  0],
                [ g3,   0,  0,   0, -g5,  0, -g3,   0,  0,   0,  g6,  0],
                [  0,  g3,  0,  g5,   0,  0,   0, -g3,  0, -g6,   0,  0],
                [  0,   0,  0,   0,   0,  0,   0,   0,  0,   0,   0,  0],
            ]) * self.material.rho * self.Ie_l * 2 / (210 * L * (1 + phi) ** 2)
            # fmt: on

        else:
            G = np.zeros((12, 12))

        return G



import re

import numpy as np
import pandas as pd
from numpy import linalg as la
from plotly import graph_objects as go
from copy import deepcopy as copy


class DataNotFoundError(Exception):
    """
    An exception indicating that the data could not be found in the file.
    """

    pass


def read_table_file(file, element, sheet_name=0, n=0, sheet_type="Model"):
    """Instantiate one or more element objects using inputs from an Excel table.

    Parameters
    ----------
    file: str
        Path to the file containing the shaft parameters.
    element: str
        Specify the type of element to be instantiated: bearing, shaft or disk.
    sheet_name: int or str, optional
        Position of the sheet in the file (starting from 0) or its name. If none is
        passed, it is assumed to be the first sheet in the file.
    n: int
        Exclusive for bearing elements, since this parameter is given outside the table
        file.
    sheet_type: str
        Exclusive for shaft elements, as they have a Model table in which more
        information can be passed, such as the material parameters.

    Returns
    -------
    A dictionary of parameters.

    Examples
    --------
    >>> import os
    >>> file_path = os.path.dirname(os.path.realpath(__file__)) + '/tests/data/shaft_si.xls'
    >>> read_table_file(file_path, "shaft", sheet_type="Model", sheet_name="Model") # doctest: +ELLIPSIS
    {'L': [0.03...
    """
    df = pd.read_excel(file, header=None, sheet_name=sheet_name)

    # Assign specific values to variables
    parameter_columns = {}
    optional_parameter_columns = {}
    header_key_word = ""
    default_dictionary = {}
    parameters = {}
    if element == "bearing":
        header_key_word = "kxx"
        parameter_columns["kxx"] = ["kxx"]
        parameter_columns["cxx"] = ["cxx"]
        optional_parameter_columns["kyy"] = ["kyy"]
        optional_parameter_columns["kxy"] = ["kxy"]
        optional_parameter_columns["kyx"] = ["kyx"]
        optional_parameter_columns["cyy"] = ["cyy"]
        optional_parameter_columns["cxy"] = ["cxy"]
        optional_parameter_columns["cyx"] = ["cyx"]
        optional_parameter_columns["frequency"] = ["frequency", "speed"]
        default_dictionary["kyy"] = None
        default_dictionary["kxy"] = 0
        default_dictionary["kyx"] = 0
        default_dictionary["cyy"] = None
        default_dictionary["cxy"] = 0
        default_dictionary["cyx"] = 0
        default_dictionary["frequency"] = None
    elif element == "shaft":
        if sheet_type == "Model":
            header_key_word = "od_left"
        else:
            header_key_word = "material"
        parameter_columns["L"] = ["length"]
        parameter_columns["idl"] = ["i_d_l", "idl", "id_left"]
        parameter_columns["odl"] = ["o_d_l", "odl", "od_left"]
        parameter_columns["idr"] = ["i_d_r", "idr", "id_right"]
        parameter_columns["odr"] = ["o_d_r", "odr", "od_right"]
        parameter_columns["material"] = ["material", "matnum"]
        optional_parameter_columns["n"] = ["n", "elemnum"]
        optional_parameter_columns["axial_force"] = [
            "axial_force",
            "axial force",
            "axial",
        ]
        optional_parameter_columns["torque"] = ["torque"]
        optional_parameter_columns["shear_effects"] = ["shear_effects", "shear effects"]
        optional_parameter_columns["rotary_inertia"] = [
            "rotary_inertia",
            "rotary inertia",
        ]
        optional_parameter_columns["gyroscopic"] = ["gyroscopic"]
        optional_parameter_columns["shear_method_calc"] = [
            "shear_method_calc",
            "shear method calc",
        ]
        default_dictionary["n"] = None
        default_dictionary["axial_force"] = 0
        default_dictionary["torque"] = 0
        default_dictionary["shear_effects"] = True
        default_dictionary["rotary_inertia"] = True
        default_dictionary["gyroscopic"] = True
        default_dictionary["shear_method_calc"] = "cowper"
    elif element == "disk":
        header_key_word = "ip"
        parameter_columns["n"] = ["unnamed: 0", "n"]
        parameter_columns["m"] = ["m", "mass"]
        parameter_columns["Id"] = ["it", "id"]
        parameter_columns["Ip"] = ["ip"]

    # Find table header and define if conversion is needed
    header_index = -1
    header_found = False
    convert_to_metric = False
    convert_to_rad_per_sec = False
    for index, row in df.iterrows():
        for i in range(0, row.size):
            if isinstance(row[i], str):
                if not header_found:
                    if row[i].lower() == header_key_word:
                        header_index = index
                        header_found = True
                if (
                    "inches" in row[i].lower()
                    or "lbm" in row[i].lower()
                    or "lb" in row[i].lower()
                ):
                    convert_to_metric = True
                if "rpm" in row[i].lower():
                    convert_to_rad_per_sec = True
                if header_found and convert_to_metric and convert_to_rad_per_sec:
                    break
        if header_found and convert_to_metric:
            break
    if not header_found:
        raise ValueError(
            "Could not find the header. Make sure the table has a header "
            "containing the names of the columns. In the case of a " + element + ", "
            "there should be a column named " + header_key_word + "."
        )

    # Get specific data from the file
    new_materials = {}
    if element == "shaft" and sheet_type == "Model":
        material_header_index = -1
        material_header_found = False
        material_header_key_word = "matno"
        for index, row in df.iterrows():
            for i in range(0, row.size):
                if isinstance(row[i], str):
                    if row[i].lower() == material_header_key_word:
                        material_header_index = index
                        material_header_found = True
                        break
            if material_header_found:
                break
        if not material_header_found:
            raise ValueError(
                "Could not find the header for the materials. Make sure the table has a header "
                "with the parameters for the materials that will be used. There should be a column "
                "named " + material_header_key_word + "."
            )
        df_material = pd.read_excel(
            file, header=material_header_index, sheet_name=sheet_name
        )
        material_name = []
        material_rho = []
        material_e = []
        material_g_s = []
        material_color = []
        for index, row in df_material.iterrows():
            if not pd.isna(row["matno"]):
                material_name.append(int(row["matno"]))
                material_rho.append(row["rhoa"])
                material_e.append(row["ea"])
                material_g_s.append(row["ga"])
                if "color" in df_material.columns:
                    if pd.isna(row["color"]):
                        material_color.append("#525252")
                    else:
                        material_color.append(row["color"])
                else:
                    material_color.append("#525252")

            else:
                break
        if convert_to_metric:
            for i in range(0, len(material_name)):
                material_rho[i] = material_rho[i] * 27679.904
                material_e[i] = material_e[i] * 6894.757
                material_g_s[i] = material_g_s[i] * 6894.757
        new_materials["matno"] = material_name
        new_materials["rhoa"] = material_rho
        new_materials["ea"] = material_e
        new_materials["ga"] = material_g_s
        new_materials["color"] = material_color

    df = pd.read_excel(file, header=header_index, sheet_name=sheet_name)
    df.columns = df.columns.str.lower()

    # Find and isolate data rows
    first_data_row_found = False
    last_data_row_found = False
    indexes_to_drop = []
    for index, row in df.iterrows():
        if (
            not first_data_row_found
            and (
                isinstance(row[header_key_word], int)
                or isinstance(row[header_key_word], float)
            )
            and not pd.isna(row[header_key_word])
        ):
            first_data_row_found = True
        elif not first_data_row_found:
            indexes_to_drop.append(index)
        elif first_data_row_found and not last_data_row_found:
            if (
                isinstance(row[header_key_word], int)
                or isinstance(row[header_key_word], float)
            ) and not pd.isna(row[header_key_word]):
                continue
            else:
                last_data_row_found = True
                indexes_to_drop.append(index)
        elif last_data_row_found:
            indexes_to_drop.append(index)
    if not first_data_row_found:
        raise DataNotFoundError(
            "Could not find the data. Make sure you have at least one row containing "
            "data below the header."
        )
    if len(indexes_to_drop) > 0:
        df = df.drop(indexes_to_drop)

    # Build parameters list
    if element == "bearing":
        parameters["n"] = n
    for key, value in parameter_columns.items():
        for name in value:
            try:
                parameters[key] = df[name].tolist()
                break
            except KeyError:
                if name == value[-1]:
                    raise ValueError(
                        "Could not find a column with one of these names: " + str(value)
                    )
                continue
    for key, value in optional_parameter_columns.items():
        for name in value:
            try:
                parameters[key] = df[name].tolist()
                break
            except KeyError:
                if name == value[-1]:
                    parameters[key] = [default_dictionary[key]] * df.shape[0]
                else:
                    continue
    if element == "shaft":
        if sheet_type == "Model":
            new_material = parameters["material"]
            for i in range(0, df.shape[0]):
                new_material[i] = "shaft_mat_" + str(int(new_material[i]))
            parameters["material"] = new_material

    # change xltrc index to python index (0 based)
    if element in ("shaft", "disk"):
        new_n = parameters["n"]
        for i in range(0, df.shape[0]):
            new_n[i] -= 1
        parameters["n"] = new_n

    if convert_to_metric:
        for i in range(0, df.shape[0]):
            if element == "bearing":
                parameters["kxx"][i] = parameters["kxx"][i] * 175.126_836_986_4
                parameters["cxx"][i] = parameters["cxx"][i] * 175.126_836_986_4
                parameters["kyy"][i] = parameters["kyy"][i] * 175.126_836_986_4
                parameters["kxy"][i] = parameters["kxy"][i] * 175.126_836_986_4
                parameters["kyx"][i] = parameters["kyx"][i] * 175.126_836_986_4
                parameters["cyy"][i] = parameters["cyy"][i] * 175.126_836_986_4
                parameters["cxy"][i] = parameters["cxy"][i] * 175.126_836_986_4
                parameters["cyx"][i] = parameters["cyx"][i] * 175.126_836_986_4
            if element == "shaft":
                parameters["L"][i] = parameters["L"][i] * 0.0254
                parameters["idl"][i] = parameters["idl"][i] * 0.0254
                parameters["odl"][i] = parameters["odl"][i] * 0.0254
                parameters["idr"][i] = parameters["idr"][i] * 0.0254
                parameters["odr"][i] = parameters["odr"][i] * 0.0254
                parameters["axial_force"][i] = (
                    parameters["axial_force"][i] * 4.448_221_615_255
                )
            elif element == "disk":
                parameters["m"][i] = parameters["m"][i] * 0.453_592_37
                parameters["Id"][i] = parameters["Id"][i] * 0.000_292_639_7
                parameters["Ip"][i] = parameters["Ip"][i] * 0.000_292_639_7
    if convert_to_rad_per_sec:
        for i in range(0, df.shape[0]):
            if element == "bearing":
                parameters["frequency"][i] = (
                    parameters["frequency"][i] * 0.104_719_755_119_7
                )
    parameters.update(new_materials)
    return parameters


def visualize_matrix(rotor, matrix, frequency=None, **kwargs):
    """Visualize global matrix.

    This function gives some visualization of a given matrix, displaying
    values on a heatmap.

    Parameters
    ----------
    rotor: rs.Rotor
        The rotor object.
    matrix: str
        String for the desired matrix.
    frequency: float, optional
        Excitation frequency. Defaults to rotor speed.
    kwargs : optional
        Additional key word arguments can be passed to change the plot layout only
        (e.g. coloraixs=dict(colorscale="Rainbow"), width=1000, height=800, ...).
        *See Plotly Python Figure Reference for more information.

    Returns
    -------
    fig : Plotly graph_objects.Figure()
        The figure object with the plot.

    Examples
    --------
    >>> import ross as rs
    >>> rotor = rs.rotor_example()

    Visualizing Mass Matrix:
    >>> fig = rs.visualize_matrix(rotor, "M", frequency=100)

    Visualizing Stiffness Matrix:
    >>> fig = rs.visualize_matrix(rotor, "K", frequency=100)

    Visualizing Gyroscopic Matrix:
    >>> fig = rs.visualize_matrix(rotor, "G")
    """
    A = np.zeros((rotor.ndof, rotor.ndof))
    # E will store element's names and contributions to the global matrix
    E = np.zeros((rotor.ndof, rotor.ndof), dtype=object)

    M, N = E.shape
    for i in range(M):
        for j in range(N):
            E[i, j] = []

    for elm in rotor.elements:
        g_dofs = list(elm.dof_global_index.values())
        l_dofs = elm.dof_local_index()
        try:
            elm_matrix = getattr(elm, matrix)(frequency)
        except TypeError:
            elm_matrix = getattr(elm, matrix)()

        A[np.ix_(g_dofs, g_dofs)] += elm_matrix

        for l0, g0 in zip(l_dofs, g_dofs):
            for l1, g1 in zip(l_dofs, g_dofs):
                if elm_matrix[l0, l1] != 0:
                    E[g0, g1].append(
                        "<br>"
                        + elm.__class__.__name__
                        + f"(n={elm.n})"
                        + f": {elm_matrix[l0, l1]:.2E}"
                    )

    # list for dofs -> ['x0', 'y0', 'alpha0', 'beta0', 'x1'...]
    dof_list = [0 for i in range(rotor.ndof)]

    for elm in rotor.elements:
        for k, v in elm.dof_global_index.items():
            dof_list[v] = k

    data = {"row": [], "col": [], "value": [], "pos_value": [], "elements": []}
    for i, dof_row in enumerate(dof_list):
        for j, dof_col in enumerate(dof_list):
            data["row"].append(i)
            data["col"].append(j)
            data["value"].append(A[i, j])
            data["pos_value"].append(abs(A[i, j]))
            data["elements"].append(E[i, j])

    dof_string_list = []
    sub = str.maketrans("0123456789", "₀₁₂₃₄₅₆₇₈₉")
    for d in dof_list:
        d = d.replace("alpha", "α")
        d = d.replace("beta", "β")
        d = d.replace("_", "")
        d = d.translate(sub)
        dof_string_list.append(d)

    x_axis = dof_string_list
    y_axis = dof_string_list[::-1]
    fig = go.Figure()
    fig.add_trace(
        go.Heatmap(
            x=x_axis,
            y=y_axis,
            z=A[::-1],
            customdata=np.array(data["elements"], dtype=object).reshape(A.shape)[::-1],
            coloraxis="coloraxis",
            hovertemplate=(
                "<b>Value: %{z:.3e}<b><br>" + "<b>Elements:<b><br> %{customdata}"
            ),
            name="<b>Matrix {}</b>".format(matrix),
        )
    )

    fig.update_layout(
        coloraxis=dict(
            cmin=np.amin(A),
            cmax=np.amax(A),
            colorscale="Rainbow",
            colorbar=dict(
                title=dict(text="<b>Value</b>", side="top"), exponentformat="power"
            ),
        ),
        **kwargs,
    )

    return fig


def convert(name):
    """Converts a CamelCase str to a underscore_case str

    Parameters
    ----------
    file: str
        Path to the file containing the shaft parameters.

    Returns
    -------
    underscore_case string

    Examples
    --------
    >>> convert('CamelCase')
    'camel_case'
    """

    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def intersection(x1, y1, x2, y2):
    """
    Intersection code from https://github.com/sukhbinder/intersection
    MIT License

    Copyright (c) 2017 Sukhbinder Singh

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

    INTERSECTIONS Intersections of curves.
       Computes the (x,y) locations where two curves intersect.  The curves
       can be broken with NaNs or have vertical segments.

    Usage:
    x, y = intersection(x1, y1, x2, y2)

    Example:
    a, b = 1, 2
    phi = np.linspace(3, 10, 100)
    x1 = a*phi - b*np.sin(phi)
    y1 = a - b*np.cos(phi)

    x2=phi
    y2=np.sin(phi)+2
    x,y=intersection(x1,y1,x2,y2)

    plt.plot(x1,y1,c='r')
    plt.plot(x2,y2,c='g')
    plt.plot(x,y,'*k')
    plt.show()

    """

    def _rect_inter_inner(x1, x2):
        n1 = x1.shape[0] - 1
        n2 = x2.shape[0] - 1
        X1 = np.c_[x1[:-1], x1[1:]]
        X2 = np.c_[x2[:-1], x2[1:]]
        S1 = np.tile(X1.min(axis=1), (n2, 1)).T
        S2 = np.tile(X2.max(axis=1), (n1, 1))
        S3 = np.tile(X1.max(axis=1), (n2, 1)).T
        S4 = np.tile(X2.min(axis=1), (n1, 1))
        return S1, S2, S3, S4

    def _rectangle_intersection_(x1, y1, x2, y2):
        S1, S2, S3, S4 = _rect_inter_inner(x1, x2)
        S5, S6, S7, S8 = _rect_inter_inner(y1, y2)

        C1 = np.less_equal(S1, S2)
        C2 = np.greater_equal(S3, S4)
        C3 = np.less_equal(S5, S6)
        C4 = np.greater_equal(S7, S8)

        ii, jj = np.nonzero(C1 & C2 & C3 & C4)
        return ii, jj

    x1 = np.asarray(x1)
    x2 = np.asarray(x2)
    y1 = np.asarray(y1)
    y2 = np.asarray(y2)

    ii, jj = _rectangle_intersection_(x1, y1, x2, y2)
    n = len(ii)

    dxy1 = np.diff(np.c_[x1, y1], axis=0)
    dxy2 = np.diff(np.c_[x2, y2], axis=0)

    T = np.zeros((4, n))
    AA = np.zeros((4, 4, n))
    AA[0:2, 2, :] = -1
    AA[2:4, 3, :] = -1
    AA[0::2, 0, :] = dxy1[ii, :].T
    AA[1::2, 1, :] = dxy2[jj, :].T

    BB = np.zeros((4, n))
    BB[0, :] = -x1[ii].ravel()
    BB[1, :] = -x2[jj].ravel()
    BB[2, :] = -y1[ii].ravel()
    BB[3, :] = -y2[jj].ravel()

    for i in range(n):
        try:
            T[:, i] = np.linalg.solve(AA[:, :, i], BB[:, i])
        except:
            T[:, i] = np.Inf

    in_range = (T[0, :] >= 0) & (T[1, :] >= 0) & (T[0, :] <= 1) & (T[1, :] <= 1)

    xy0 = T[2:, in_range]
    xy0 = xy0.T
    return xy0[:, 0], xy0[:, 1]


def get_data_from_figure(fig):
    """Get data from a Plotly ccatter plot (XY) and convert to a pd.DataFrame.

    This function takes a go.Figure() object from Plotly and converts the data to a
    DataFrame from Pandas. It works only for Scatter Plots (XY) and does no support
    subplots or polar plots.

    This function is suitable to the following analyzes:
        - run_freq_response()
        - run_unbalance_response()
        - run_time_response()

    Parameters
    ----------
    fig : go.Figure
        A Plotly figure generated by go.Figure() command.

    Returns
    -------
    df : pd.DataFrame
        A DataFrame with the fig.data.

    Examples
    --------
    >>> import ross as rs
    >>> rotor = rs.rotor_example()

    Get post-processed data from an unbalance response
    >>> resp = rotor.run_unbalance_response(
    ...     3, 0.001, 0.0, np.linspace(0, 1000, 101)
    ... )
    >>> fig = resp.plot_magnitude(probe=[rs.Probe(3, 0.0, tag="probe1"), rs.Probe(3, np.pi/2, tag="probe2")])
    >>> df = rs.get_data_from_figure(fig)

    Use the probe tag to navigate through pandas data
    Index 0 for frequency array
    >>> df["probe1"].values[0] # doctest: +ELLIPSIS
    array([   0.,   10.,   20.,   30.,...

    Index 1 for amplitude array
    >>> df["probe1"].values[1] # doctest: +ELLIPSIS
    array([0.00000000e+00,...

    Or use "iloc" to obtain the desired array from pandas
    >>> df.iloc[1, 0] # doctest: +ELLIPSIS
    array([0.00000000e+00, 1.6057...
    """
    dict_data = {data["name"]: {} for data in fig.data}

    xaxis_label = fig.layout.xaxis.title.text
    yaxis_label = fig.layout.yaxis.title.text

    for i, data in enumerate(fig.data):
        dict_data[data["name"]][xaxis_label] = data.x
        dict_data[data["name"]][yaxis_label] = data.y

    df = pd.DataFrame(dict_data)

    return df


def newmark(func, t, y_size, **options):
    """Transient solution of the dynamic behavior of the system.

    Perform numerical integration using the Newmark method with Newton-Raphson
    iterations of the generic equation of motion:
    M * y'' + C * y' + K * y = RHS(t, y)

    Parameters
    ----------
    func : callable
        A function that calculates the system matrices and right-hand side (RHS) vector at each
        time step. It should take at least one argument `(step, dt=None, y=None, ydot=None, y2dot=None)`
        and return a tuple `(M, C, K, RHS)`, where `step` is a scalar int related to the current time
        step, `dt` is the current time step in seconds, `y` is a ndarray of current state of the system,
        `ydot` and `y2dot` are its first and second time derivatives. `M`, `C`, `K` are ndarrays with
        `np.shape(M) = (y_size, y_size)` and `RHS` is a ndarray with `len(RHS) = y_size`.
    t : array_like
        Time array.
    y_size : int
        Size of the state vector.
    **options
        Options passed for controlling the integration parameters. All options available are
        listed below.
    gamma : float, optional
        Parameter of the integration algorithm related to the velocity interpolation equation.
        Default is 0.5.
    beta : float, optional
        Parameter of the integration algorithm related to the displacement interpolation equation.
        Default is 0.25.
    tol : float, optional
        Convergence tolerance for the Newton-Raphson iterations. Default is 1e-6.
    progress_interval : float, optional
        Time interval at which progress is printed. Default is to not show progress.

    Returns
    -------
    yout : ndarray
        System response. It is an array containing the state variables at each time step of `t` with
        `np.shape(yout) = (len(t), y_size)`

    References
    ----------
    Newmark, N. M. (1959). A method of computation for structural dynamics.
    Journal of the Engineering Mechanics Division, 85(3), 67-94.

    Examples
    --------
    >>> import ross as rs
    >>> rotor = rs.rotor_example()
    >>> size = 10000
    >>> node = 3
    >>> speed = 500.0
    >>> accel = 0.0
    >>> t = np.linspace(0, 10, size)
    >>> F = np.zeros((size, rotor.ndof))
    >>> F[:, rotor.number_dof * node] = 10 * np.cos(2 * t)
    >>> F[:, rotor.number_dof * node + 1] = 10 * np.sin(2 * t)
    >>> M = rotor.M(speed)
    >>> C1 = rotor.C(speed)
    >>> K1 = rotor.K(speed)
    >>> C2 = rotor.G()
    >>> K2 = rotor.Ksdt()
    >>> rotor_system = lambda i, **state: (M, C1 + C2 * speed, K1 + K2 * accel, F[i, :])
    >>> yout = newmark(rotor_system, t, rotor.ndof)
    >>> dof = 13
    >>> yout[:, dof] # doctest: +ELLIPSIS
    array([0.00000000e+00, 8.49140057e-09, 4.34296767e-08, ...,
           1.16148468e-05, 1.16492353e-05, 1.16859622e-05])
    """

    gamma = options.get("gamma", 0.5)
    beta = options.get("beta", 0.25)
    tol = options.get("tol", 1e-6)
    progress_interval = options.get("progress_interval", t[-1] + 1)

    n_steps = len(t)
    ny = y_size

    y0 = np.zeros(ny)
    ydot0 = np.zeros(ny)
    y2dot0 = np.zeros(ny)

    yout = np.full((n_steps, ny), 1e-38, dtype=t.dtype)
    yout[0, :] = y0

    for step in range(1, n_steps):
        aux = round(t[step] / progress_interval, 9)
        if aux - int(aux) == 0:
            print(f"Time: {t[step]:.6f} seconds")

        dt = t[step] - t[step - 1]

        M, C, K, RHS = func(step, dt=dt, y=y0, ydot=ydot0, y2dot=y2dot0)

        y2dot = np.zeros(ny)
        ydot = ydot0 + y2dot0 * (1 - gamma) * dt
        y = y0 + ydot0 * dt + y2dot0 * (0.5 - beta) * (dt**2)

        res = RHS - (K @ y + C @ ydot) - M @ y2dot
        nr_iter = 0

        while la.norm(res) >= tol:
            nr_iter += 1
            if nr_iter > 1e5:
                raise Warning(
                    "The Newton-Raphson algorithm is taking a long time to converge."
                )

            dy2dot = la.solve(M + C * gamma * dt + K * beta * (dt**2), res)

            y2dot += dy2dot
            ydot += dy2dot * gamma * dt
            y += dy2dot * beta * (dt**2)

            res = RHS - (K @ y + C @ ydot) - M @ y2dot

        y0 = y
        ydot0 = ydot
        y2dot0 = y2dot

        yout[step, :] = y

    return yout


def assemble_C_K_matrices(elements, C0, K0, *args):
    """Assemble damping and stiffness matrices considering
    specified elements a rotor.

    Parameters
    ----------
    elements : array_like
        List of elements of the rotor.
    C0 : ndarray
        Initial damping matrix.
    K0 : ndarray
        Initial stiffness matrix.
    args : float, optional
        Any additional argument that can be passed to the element C or K methods.

    Returns
    -------
    C0 : ndarray
        Initial damping matrix incorporating the specified elements.
    K0 : ndarray
        Initial stiffness matrix incorporating the specified elements.

    Examples
    --------
    >>> import ross as rs
    >>> rotor = rs.rotor_example()
    >>> elements_without_bearing = [
    ...     *rotor.shaft_elements,
    ...     *rotor.disk_elements,
    ...     *rotor.point_mass_elements
    ... ]
    >>> C0 = np.zeros((rotor.ndof, rotor.ndof))
    >>> K0 = np.zeros((rotor.ndof, rotor.ndof))
    >>> C1, K1 = assemble_C_K_matrices(elements_without_bearing, C0, K0)
    >>> C, K = assemble_C_K_matrices(rotor.bearing_elements, C1, K1, 0)
    >>> C[:4, :4]
    array([[0., 0., 0., 0.],
           [0., 0., 0., 0.],
           [0., 0., 0., 0.],
           [0., 0., 0., 0.]])
    >>> np.round(K[:4, :4]/1e6)
    array([[47.,  0.,  0.,  6.],
           [ 0., 46., -6.,  0.],
           [ 0., -6.,  1.,  0.],
           [ 6.,  0.,  0.,  1.]])
    """

    for elm in elements:
        dofs = list(elm.dof_global_index.values())
        try:
            C0[np.ix_(dofs, dofs)] += elm.C(*args)
        except:
            C0[np.ix_(dofs, dofs)] += elm.C()
        try:
            K0[np.ix_(dofs, dofs)] += elm.K(*args)
        except:
            K0[np.ix_(dofs, dofs)] += elm.K()

    return C0, K0


def remove_dofs(matrix, dofs=None):
    """Removes specified degrees of freedom from the given matrix.

    By default, this function takes a 6 dof model matrix and removes the axial and torsional dofs,
    resulting in a 4 dof model matrix.

    Parameters
    ----------
    matrix: ndarray
        The original matrix to process.
    dofs: list
        List of indices representing dofs to be removed. Default is None, but internally it considers
        the axial and torsional dofs of a 6 dof model.

    Returns
    -------
    new_matrix: ndarray
        The modified matrix with the removed dofs.

    Examples
    --------
    >>> import ross as rs
    >>> rotor = rs.rotor_example_6dof()
    >>> n_nodes = rotor.nodes[-1] + 1
    >>> M_6dof = rotor.M()
    >>> M_4dof = remove_dofs(M_6dof)
    >>> M_6dof.shape
    (42, 42)
    >>> len(M_6dof) == n_nodes * 6
    True
    >>> M_4dof.shape
    (28, 28)
    >>> len(M_4dof) == n_nodes * 4
    True
    """
    if dofs is None:
        dofs = np.arange(2, len(matrix), 3)

    new_matrix = np.delete(np.delete(matrix, dofs, axis=0), dofs, axis=1)

    return new_matrix


def convert_6dof_to_4dof(rotor):
    """Convert a 6 dof rotor model to a 4 dof model.

    This function takes a 6 dof rotor model and modifies it by removing the axial and
    torsional dofs. It adjusts the corresponding matrix methods to reflect this change.

    Parameters
    ----------
    rotor: rs.Rotor
        The rotor object of 6 dof model.

    Returns
    -------
    new_rotor: rs.Rotor
        The rotor object modified.

    Examples
    --------
    >>> import ross as rs
    >>> rotor = rs.rotor_example_6dof()
    >>> rotor_mod = convert_6dof_to_4dof(rotor)
    >>> n_nodes = rotor.nodes[-1] + 1
    >>> M_6dof = rotor.M()
    >>> M_4dof = rotor_mod.M()
    >>> M_6dof.shape
    (42, 42)
    >>> len(M_6dof) == n_nodes * 6
    True
    >>> M_4dof.shape
    (28, 28)
    >>> len(M_4dof) == n_nodes * 4
    True
    """
    # Copy the rotor object
    new_rotor = copy(rotor)

    # Modify matrix methods to get 4 dof matrices
    new_rotor.M = lambda frequency=None, synchronous=False: remove_dofs(
        rotor.M(frequency=frequency, synchronous=synchronous)
    )
    new_rotor.K = lambda frequency: remove_dofs(rotor.K(frequency))
    new_rotor.Ksdt = lambda: remove_dofs(rotor.Ksdt())
    new_rotor.C = lambda frequency: remove_dofs(rotor.C(frequency))
    new_rotor.G = lambda: remove_dofs(rotor.G())

    # Update number of dofs
    new_rotor.number_dof = 4
    new_rotor.ndof = len(new_rotor.M())

    return new_rotor



import os
import pickle
from pathlib import Path
from tempfile import tempdir

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_almost_equal

from ross.disk_element import DiskElement, DiskElement6DoF
from ross.materials import steel


@pytest.fixture
def disk():
    return DiskElement(0, 0.07, 0.05, 0.32956)


def test_index(disk):
    assert disk.dof_local_index().x_0 == 0
    assert disk.dof_local_index().y_0 == 1
    assert disk.dof_local_index().alpha_0 == 2
    assert disk.dof_local_index().beta_0 == 3


def test_mass_matrix_disk(disk):
    # fmt: off
    Md1 = np.array([[0.07, 0., 0., 0.],
                    [0., 0.07, 0., 0.],
                    [0., 0., 0.05, 0.],
                    [0., 0., 0., 0.05]])
    # fmt: on

    assert_almost_equal(disk.M(), Md1, decimal=5)


def test_gyroscopic_matrix_disk(disk):
    # fmt: off
    Gd1 = np.array([[0., 0., 0., 0.],
                    [0., 0., 0., 0.],
                    [0., 0., 0., 0.32956],
                    [0., 0., -0.32956, 0.]])
    # fmt: on

    assert_almost_equal(disk.G(), Gd1, decimal=5)


@pytest.fixture
def disk_from_geometry():
    return DiskElement.from_geometry(0, steel, 0.07, 0.05, 0.28)


def test_pickle(disk):
    disk_pickled = pickle.loads(pickle.dumps(disk))
    assert disk == disk_pickled


def test_save_load(disk):
    file = Path(tempdir) / "disk.toml"
    disk.save(file)
    disk_loaded = DiskElement.load(file)
    assert disk == disk_loaded


def test_disk_equality(disk_from_geometry, disk):
    assert disk != disk_from_geometry


def test_mass_matrix_disk1(disk_from_geometry):
    # fmt: off
    Md1 = np.array([[ 32.58973,   0.     ,   0.     ,   0.     ],
                    [  0.     ,  32.58973,   0.     ,   0.     ],
                    [  0.     ,   0.     ,   0.17809,   0.     ],
                    [  0.     ,   0.     ,   0.     ,   0.17809]])
    # fmt: on
    assert_almost_equal(disk_from_geometry.M(), Md1, decimal=5)


def test_gyroscopic_matrix_disk1(disk_from_geometry):
    # fmt: off
    Gd1 = np.array([[ 0.     ,  0.     ,  0.     ,  0.     ],
                    [ 0.     ,  0.     ,  0.     ,  0.     ],
                    [ 0.     ,  0.     ,  0.     ,  0.32956],
                    [ 0.     ,  0.     , -0.32956,  0.     ]])
    # fmt: on
    assert_almost_equal(disk_from_geometry.G(), Gd1, decimal=5)


def test_from_table():
    for file_name in ["/data/shaft_us.xls", "/data/shaft_si.xls"]:
        file_name = os.path.dirname(os.path.realpath(__file__)) + file_name
        disks = DiskElement.from_table(file_name, sheet_name="More")

        assert_allclose(disks[1].m, 6.90999178227835)
        assert_allclose(disks[1].Ip, 0.0469996988106328, atol=1.6e-07)
        assert_allclose(disks[1].Id, 0.0249998397928898, atol=1.6e-07)


@pytest.fixture
def disk_6dof():
    return DiskElement6DoF(0, 32.58973, 0.17809, 0.32956)


def test_index_6dof(disk_6dof):
    assert disk_6dof.dof_local_index().x_0 == 0
    assert disk_6dof.dof_local_index().y_0 == 1
    assert disk_6dof.dof_local_index().z_0 == 2
    assert disk_6dof.dof_local_index().alpha_0 == 3
    assert disk_6dof.dof_local_index().beta_0 == 4
    assert disk_6dof.dof_local_index().theta_0 == 5


def test_mass_matrix_disk_6dof(disk_6dof):
    # fmt: off
    Md1 = np.array([[32.58973, 0.,       0.,       0.,      0.,      0.   ],
                    [0.,       32.58973, 0.,       0.,      0.,      0.   ],
                    [0.,       0.,       32.58973, 0.,      0.,      0.   ],
                    [0.,       0.,       0.,       0.17809, 0.,      0.   ],
                    [0.,       0.,       0.,       0.,      0.17809, 0.   ],
                    [0.,       0.,       0.,       0,       0.,      0.32956]])
    # fmt: on
    assert_almost_equal(disk_6dof.M(), Md1, decimal=5)


def test_gyroscopic_matrix_disk_6dof(disk_6dof):
    # fmt: off
    Gd1 = np.array([
            [ 0, 0, 0,  0,         0,       0],
            [ 0, 0, 0,  0,         0,       0],
            [ 0, 0, 0,  0,         0,       0],
            [ 0, 0, 0,  0,         0.32956, 0],
            [ 0, 0, 0, -0.32956,   0,       0],
            [ 0, 0, 0,  0,         0,       0]])
    # fmt: on

    assert_almost_equal(disk_6dof.G(), Gd1, decimal=5)


@pytest.fixture
def disk_from_geometry_6dof():
    return DiskElement6DoF.from_geometry(0, steel, 0.07, 0.05, 0.28)


def test_mass_matrix_disk1_6dof(disk_from_geometry_6dof):
    # fmt: off
    Md1 = np.array([[32.58973, 0.,       0.,       0.,      0.,      0.   ],
                    [0.,       32.58973, 0.,       0.,      0.,      0.   ],
                    [0.,       0.,       32.58973, 0.,      0.,      0.   ],
                    [0.,       0.,       0.,       0.17809, 0.,      0.   ],
                    [0.,       0.,       0.,       0.,      0.17809, 0.   ],
                    [0.,       0.,       0.,       0,       0.,      0.32956]])
    # fmt: on

    M_class = disk_from_geometry_6dof.M()
    assert_almost_equal(M_class, Md1, decimal=5)


def test_gyroscopic_matrix_disk1_6dof(disk_from_geometry_6dof):
    # fmt: off
    Gd1 = np.array([
            [ 0, 0, 0,  0,         0,       0],
            [ 0, 0, 0,  0,         0,       0],
            [ 0, 0, 0,  0,         0,       0],
            [ 0, 0, 0,  0,         0.32956, 0],
            [ 0, 0, 0, -0.32956,   0,       0],
            [ 0, 0, 0,  0,         0,       0]])
    # fmt: on
    assert_almost_equal(disk_from_geometry_6dof.G(), Gd1, decimal=5)


def test_from_table_6dof():
    for file_name in ["/data/shaft_us.xls", "/data/shaft_si.xls"]:
        file_name = os.path.dirname(os.path.realpath(__file__)) + file_name
        disks = DiskElement6DoF.from_table(file_name, sheet_name="More")

        assert_allclose(disks[1].m, 6.90999178227835)
        assert_allclose(disks[1].Ip, 0.0469996988106328, atol=1.6e-07)
        assert_allclose(disks[1].Id, 0.0249998397928898, atol=1.6e-07)



import pickle
from collections import namedtuple

import pytest
from numpy.testing import assert_allclose

from ross.units import Q_, check_units


def test_new_units_loaded():
    speed = Q_(1, "RPM")
    assert speed.m == 1
    # check if h is hour instead of planck constant
    v = Q_(3600, "m/h")
    assert v.to("m/s").m == 1


def test_units_pickle():
    speed = Q_(1, "RPM")
    speed_pickled = pickle.loads(pickle.dumps(speed))
    assert speed == speed_pickled


# each possible argument
Argument = namedtuple(
    "Argument", ["single_value", "si_unit", "other_unit", "expected_converted_value"]
)
arguments = {
    "E": Argument(1, "N/m²", "lbf/in²", 6894.7572931683635),
    "G_s": Argument(1, "N/m²", "lbf/in²", 6894.7572931683635),
    "rho": Argument(1, "kg/m³", "lb/foot**3", 16.01846337396014),
    "L": Argument(1, "meter", "inches", 0.0254),
    "idl": Argument(1, "meter", "inches", 0.0254),
    "idr": Argument(1, "meter", "inches", 0.0254),
    "odl": Argument(1, "meter", "inches", 0.0254),
    "odr": Argument(1, "meter", "inches", 0.0254),
    "speed": Argument(1, "rad/s", "RPM", 0.10471975511965977),
    "frequency": Argument(1, "rad/s", "RPM", 0.10471975511965977),
    "frequency_range": Argument(
        (1, 1), "rad/s", "RPM", (0.10471975511965977, 0.10471975511965977)
    ),
    "m": Argument(1, "kg", "lb", 0.4535923700000001),
    "mx": Argument(1, "kg", "lb", 0.4535923700000001),
    "my": Argument(1, "kg", "lb", 0.4535923700000001),
    "Ip": Argument(1, "kg*m²", "lb*in²", 0.0002926396534292),
    "Id": Argument(1, "kg*m²", "lb*in²", 0.0002926396534292),
    "width": Argument(1, "meter", "inches", 0.0254),
    "depth": Argument(1, "meter", "inches", 0.0254),
    "pitch": Argument(1, "meter", "inches", 0.0254),
    "height": Argument(1, "meter", "inches", 0.0254),
    "shaft_radius": Argument(1, "meter", "inches", 0.0254),
    "radial_clearance": Argument(1, "meter", "inches", 0.0254),
    "i_d": Argument(1, "meter", "inches", 0.0254),
    "o_d": Argument(1, "meter", "inches", 0.0254),
    "unbalance_magnitude": Argument(1, "kg*m", "lb*in", 0.011521246198000002),
    "unbalance_phase": Argument(1, "rad", "deg", 0.017453292519943295),
    "inlet_pressure": Argument(1, "pascal", "kgf/cm²", 98066.5),
    "outlet_pressure": Argument(1, "pascal", "kgf/cm²", 98066.5),
    "inlet_temperature": Argument(1, "degK", "degC", 274.15),
    "inlet_swirl_velocity": Argument(1, "m/s", "ft/s", 0.3048),
    "pad_thickness": Argument(1, "meter", "inches", 0.0254),
    "kxx": Argument(1, "N/m", "lbf/in", 175.12683524647645),
    "kxy": Argument(1, "N/m", "lbf/in", 175.12683524647645),
    "kxz": Argument(1, "N/m", "lbf/in", 175.12683524647645),
    "kyx": Argument(1, "N/m", "lbf/in", 175.12683524647645),
    "kyy": Argument(1, "N/m", "lbf/in", 175.12683524647645),
    "kyz": Argument(1, "N/m", "lbf/in", 175.12683524647645),
    "kzx": Argument(1, "N/m", "lbf/in", 175.12683524647645),
    "kzy": Argument(1, "N/m", "lbf/in", 175.12683524647645),
    "kzz": Argument(1, "N/m", "lbf/in", 175.12683524647645),
    "cxx": Argument(1, "N*s/m", "lbf*s/in", 175.12683524647645),
    "cxy": Argument(1, "N*s/m", "lbf*s/in", 175.12683524647645),
    "cxz": Argument(1, "N*s/m", "lbf*s/in", 175.12683524647645),
    "cyx": Argument(1, "N*s/m", "lbf*s/in", 175.12683524647645),
    "cyy": Argument(1, "N*s/m", "lbf*s/in", 175.12683524647645),
    "cyz": Argument(1, "N*s/m", "lbf*s/in", 175.12683524647645),
    "czx": Argument(1, "N*s/m", "lbf*s/in", 175.12683524647645),
    "czy": Argument(1, "N*s/m", "lbf*s/in", 175.12683524647645),
    "czz": Argument(1, "N*s/m", "lbf*s/in", 175.12683524647645),
    "viscosity": Argument(1, "Pa*s", "lb/ft/s", 1.488164),
    "weight": Argument(1, "N", "lbf", 4.4482216152605005),
    "load": Argument(1, "N", "lbf", 4.4482216152605005),
    "force": Argument(1, "N", "lbf", 4.4482216152605005),
    "torque": Argument(1, "N*m", "lbf*ft", 1.3558179483314006),
}


@pytest.fixture
def auxiliary_function():
    @check_units
    def func(
        E,
        G_s,
        rho,
        L,
        idl,
        idr,
        odl,
        odr,
        speed,
        frequency,
        frequency_range,
        m,
        mx,
        my,
        Ip,
        Id,
        width,
        depth,
        pitch,
        height,
        shaft_radius,
        radial_clearance,
        i_d,
        o_d,
        unbalance_magnitude,
        unbalance_phase,
        inlet_pressure,
        outlet_pressure,
        inlet_temperature,
        inlet_swirl_velocity,
        pad_thickness,
        kxx,
        kxy,
        kxz,
        kyx,
        kyy,
        kyz,
        kzx,
        kzy,
        kzz,
        cxx,
        cxy,
        cxz,
        cyx,
        cyy,
        cyz,
        czx,
        czy,
        czz,
        viscosity,
        weight,
        load,
        force,
        torque,
    ):
        return (
            E,
            G_s,
            rho,
            L,
            idl,
            idr,
            odl,
            odr,
            speed,
            frequency,
            frequency_range,
            m,
            mx,
            my,
            Ip,
            Id,
            width,
            depth,
            pitch,
            height,
            shaft_radius,
            radial_clearance,
            i_d,
            o_d,
            unbalance_magnitude,
            unbalance_phase,
            inlet_pressure,
            outlet_pressure,
            inlet_temperature,
            inlet_swirl_velocity,
            pad_thickness,
            kxx,
            kxy,
            kxz,
            kyx,
            kyy,
            kyz,
            kzx,
            kzy,
            kzz,
            cxx,
            cxy,
            cxz,
            cyx,
            cyy,
            cyz,
            czx,
            czy,
            czz,
            viscosity,
            weight,
            load,
            force,
            torque,
        )

    return func


def test_units(auxiliary_function):
    results = auxiliary_function(**{k: v.single_value for k, v in arguments.items()})
    results_dict = {k: v for k, v in zip(arguments.keys(), results)}
    for arg, actual in results_dict.items():
        assert_allclose(actual, arguments[arg].single_value)


def test_unit_Q_(auxiliary_function):
    kwargs = {k: Q_(v.single_value, v.si_unit) for k, v in arguments.items()}
    results = auxiliary_function(**kwargs)
    results_dict = {k: v for k, v in zip(arguments.keys(), results)}
    for arg, actual in results_dict.items():
        assert_allclose(actual, arguments[arg].single_value)


def test_unit_Q_conversion(auxiliary_function):
    kwargs = {k: Q_(v.single_value, v.other_unit) for k, v in arguments.items()}
    results = auxiliary_function(**kwargs)
    results_dict = {k: v for k, v in zip(arguments.keys(), results)}
    for arg, actual in results_dict.items():
        assert_allclose(actual, arguments[arg].expected_converted_value)
